#ifndef __BA_ASSERT_H__
#define __BA_ASSERT_H__

#include <ba_config.h>
#include <stdlib.h>

// XXX violation of naming convention...  more proper to do something like
// BA_ASSERT_FAIL() or just remove BA_PANIC and replace it with
// BA_ASSERT(0) in code.  not a big deal...  same with BA_PRINT_LOCATION().

#define BA_PANIC() { \
    ba_print_error("Fatal error in %s(), %s:%d.\n", __func__, __FILE__, \
      __LINE__); \
    ba_assert_exit_badly(); \
}

#define BA_PRINT_LOCATION() { \
    ba_print_info("BA_PRINT_LOCATION(): %s(), %s:%d.\n", __func__, __FILE__, \
      __LINE__); \
}

#define BA_ASSERT(_x_) { \
    if (!(_x_)) { \
        ba_print_error("Assertion check in %s(), %s:%d.\n", __func__, \
          __FILE__, __LINE__); \
        ba_print_error("Assertion failed: %s\n", #_x_); \
        ba_assert_exit_badly(); \
    } \
}

#define BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(_a_, _b_) { \
    if(ba_types_real_greater_than_or_equal(_a_, _b_) == BA_FALSE) { \
        ba_print_error("Assertion check in %s(), %s:%d.\n", __func__, \
          __FILE__, __LINE__); \
        ba_print_error("Assertion failed: " BA_REAL_FMT " >= " BA_REAL_FMT \
                       "\n", _a_, _b_); \
        ba_assert_exit_badly(); \
    } \
}

#define BA_ASSERT_REAL_LESS_THAN_OR_EQUAL(_a_, _b_) { \
    if(ba_types_real_less_than_or_equal(_a_, _b_) == BA_FALSE) { \
        ba_print_error("Assertion check in %s(), %s:%d.\n", __func__, \
          __FILE__, __LINE__); \
        ba_print_error("Assertion failed: " BA_REAL_FMT " <= " BA_REAL_FMT \
                       "\n", _a_, _b_); \
        ba_assert_exit_badly(); \
    } \
}

#define BA_ASSERT_REAL_EQUAL(_a_, _b_) { \
    if(ba_types_real_equal(_a_, _b_) == BA_FALSE) { \
        ba_print_error("Assertion check in %s(), %s:%d.\n", __func__, \
          __FILE__, __LINE__); \
        ba_print_error("Assertion failed: " BA_REAL_FMT " == " BA_REAL_FMT \
                       "\n", _a_, _b_); \
        ba_assert_exit_badly(); \
    } \
}

extern void ba_assert_exit_badly(void);

extern ba_status ba_assert_init(void);
extern ba_status ba_assert_destroy(void);
extern void ba_assert_destroy_from_hook(void *unused);

#endif /* !__BA_ASSERT_H__ */
