#include <ba_config.h>
#include <ba_sim.h>

#define EVENTS_DEBUG_PRINT 0

// ugly, but i see no way around it...  i'm replicating some fields of
// a ba_event_t, which is stuff usually hidden in ba_event.c, not
// ba_events.c.
typedef struct {
    ba_event_kind_t kind;
    ba_real action_abs_time;
    ba_event_callback_t *callback;
    void *closure;

    ba_bool found_earlier;
    ba_bool found_later;
    ba_real entered_abs_time;
} replace_later_t;

static gint event_time_compare(gconstpointer event_a_in,
			       gconstpointer event_b_in);
static void create_and_insert_event(ba_event_kind_t kind,
				    ba_real entered_abs_time,
				    ba_real action_abs_time,
				    ba_event_callback_t callback,
				    void *closure);
static replace_later_t *replace_later_new(ba_event_kind_t kind,
					  ba_real action_abs_time,
					  ba_event_callback_t callback,
					  void *closure);
static void replace_later_delete(replace_later_t *replace_later);
static gint find_later_events(gconstpointer event_in,
			      gconstpointer replace_later_in);
static void do_one_event(ba_event_t *event);

static const ba_real virtual_clock_initial_abs_time = 0.0;

static GSList *events;
static ba_real virtual_clock_abs_time;

static ba_id_gen_t *id_gen;

extern ba_real
ba_events_virtual_clock_initial_abs_time_get(void)
{
    return virtual_clock_initial_abs_time;
}

extern ba_real
ba_events_virtual_clock_abs_time_get(void)
{
    return virtual_clock_abs_time;
}

static gint
event_time_compare(gconstpointer event_a_in,
		   gconstpointer event_b_in)
{
    BA_ASSERT(event_a_in && event_b_in);
    ba_event_t *event_a, *event_b;
    ba_real action_abs_time_a, action_abs_time_b;

    event_a = (ba_event_t *)event_a_in;
    event_b = (ba_event_t *)event_b_in;

    action_abs_time_a = ba_event_action_abs_time_get(event_a);
    action_abs_time_b = ba_event_action_abs_time_get(event_b);

    // can't just return a - b, since this function returns an int, not a real.
    /*    if(ba_types_real_greater_than_or_equal(action_abs_time_a,
	  action_abs_time_b)) {*/

    // forget rounding error.  assume that numbers are perfect.  now,
    // if the event we want to add, which is event_a, has an equal
    // time to b, which is already on the list, put event_a second.
    // helpful in getting the order right for all those cases where i
    // mark an event as BA_EVENTS_NOW.
    if(action_abs_time_a >= action_abs_time_b) {
	return 1;
    } else {
	return -1;
    }
}

static void
create_and_insert_event(ba_event_kind_t kind,
			ba_real entered_abs_time,
			ba_real action_abs_time,
			ba_event_callback_t callback,
			void *closure)
{
    ba_id_t id;
    ba_event_t *event;

    id = ba_id_gen_next(id_gen);

    event = ba_event_new(id, kind, entered_abs_time, action_abs_time, callback,
			 closure);
    events = g_slist_insert_sorted(events, event, event_time_compare);

#if EVENTS_DEBUG_PRINT
    ba_event_print(event);
#endif
}

extern void
ba_events_add_event(ba_event_kind_t kind,
		    ba_real action_delta_time,
		    ba_event_callback_t callback,
		    void *closure)
{
    BA_ASSERT(callback);
    ba_real entered_abs_time, action_abs_time;

    action_abs_time = virtual_clock_abs_time + action_delta_time;
    entered_abs_time = virtual_clock_abs_time;

#if EVENTS_DEBUG_PRINT
    ba_print_info("ADDED:\n");
#endif

    create_and_insert_event(kind, entered_abs_time, action_abs_time, callback,
			    closure);
}

extern uint32_t
ba_events_remove_events(void *data,
			GCompareFunc func)
{
    // data might be 0
    BA_ASSERT(func);
    uint32_t count;
    GSList *elem;
    ba_event_t *event;

    count = 0;
    while(1) {
	elem = g_slist_find_custom(events, data, func);
	if(elem) {
	    event = (ba_event_t *)elem->data;
	    events = g_slist_remove(events, event);

#if EVENTS_DEBUG_PRINT
	    ba_print_info("REMOVED:\n");
	    ba_event_print(event);
#endif

	    ba_event_delete(event);
	    count++;
	} else {
	    break;
	}
    }

    return count;
}

static replace_later_t *
replace_later_new(ba_event_kind_t kind,
		  ba_real action_abs_time,
		  ba_event_callback_t callback,
		  void *closure)
{
    BA_ASSERT(callback);
    replace_later_t *replace_later;

    replace_later = ba_mem_malloc(sizeof(replace_later_t));

    replace_later->kind = kind;
    replace_later->action_abs_time = action_abs_time;
    replace_later->callback = callback;
    replace_later->closure = closure;

    replace_later->found_earlier = BA_FALSE;
    replace_later->found_later = BA_FALSE;

    return replace_later;
}

static void
replace_later_delete(replace_later_t *replace_later)
{
    BA_ASSERT(replace_later);

    ba_mem_free(replace_later);
}

static gint
find_later_events(gconstpointer event_in,
		  gconstpointer replace_later_in)
{
    BA_ASSERT(event_in && replace_later_in);
    ba_event_t *event;
    replace_later_t *replace_later;
    
    event = (ba_event_t *)event_in;
    replace_later = (replace_later_t *)replace_later_in;

    if(replace_later->found_earlier) {
	// we found out we shouldn't be messing around here.  bow out
	// repeatedly.
	return 1;
    }
    
    if(replace_later->kind != ba_event_kind_get(event) ||
       replace_later->callback != ba_event_callback_get(event) ||
       replace_later->closure != ba_event_closure_get(event)) {
	// not the kind of event we are interested in...
	return 1;
    }
    
    if(ba_types_real_greater_than_or_equal(replace_later->action_abs_time,
					   ba_event_action_abs_time_get
					   (event))) {
	replace_later->found_earlier = BA_TRUE;
	// aha, there's someone before us.  bow out...
	return 1;
    }
    
    // ok, we found someone that looks like us and we come strictly
    // before them.  so, poof they go.

    // there should only be one before us...
    BA_ASSERT(replace_later->found_later == BA_FALSE);

    replace_later->found_later = BA_TRUE;
    replace_later->entered_abs_time = ba_event_entered_abs_time_get(event);

    return 0;
}

extern void
ba_events_replace_later_event(ba_event_kind_t kind,
			      ba_real action_delta_time,
			      ba_event_callback_t callback,
			      void *closure)
{
    BA_ASSERT(callback);
    replace_later_t *replace_later;
    ba_bool found_earlier;
    ba_bool found_later;
    ba_real entered_abs_time, replaced_entered_abs_time, action_abs_time;

    action_abs_time = virtual_clock_abs_time + action_delta_time;

    replace_later = replace_later_new(kind, action_abs_time, callback,
				      closure);
    ba_events_remove_events(replace_later, find_later_events);

    found_earlier = replace_later->found_earlier;
    found_later = replace_later->found_later;
    replaced_entered_abs_time = replace_later->entered_abs_time;
    replace_later_delete(replace_later);

    if(found_earlier) {
	return;
    }

    if(found_later) {
	entered_abs_time = replaced_entered_abs_time;
    } else {
	entered_abs_time = virtual_clock_abs_time;
    }

#if EVENTS_DEBUG_PRINT
    ba_print_info("ADDED (possibly a replacement):\n");
#endif

    create_and_insert_event(kind, entered_abs_time, action_abs_time, callback,
			    closure);
}

extern void
ba_events_simulate(void)
{
    ba_event_t *event;

    while(events) {
	event = (ba_event_t *)events->data;
	BA_ASSERT(event);

#if EVENTS_DEBUG_PRINT
	ba_print_info("ABOUT TO DO:\n");
	ba_event_print(event);
#endif

	do_one_event(event);
    }

    ba_report_events_virtual_clock_start_end(virtual_clock_initial_abs_time,
					     virtual_clock_abs_time);
}

static void
do_one_event(ba_event_t *event)
{
    ba_real action_abs_time;

    events = g_slist_remove(events, event);
    
    action_abs_time = ba_event_action_abs_time_get(event);
    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(action_abs_time,
					 virtual_clock_abs_time);
    
    if(action_abs_time >= virtual_clock_abs_time) {
	// this `if' is here to ensure that time doesn't go
	// backward in case rounding error has placed
	// action_abs_time slightly under virtual_clock_abs_time.
	// but is it really bad for time to go backward by
	// BA_REAL_TOLERANCE or less?  XXX
	
	virtual_clock_abs_time = action_abs_time;
    }
    
    ba_event_invoke_callback(event);

    ba_event_delete(event);
}

extern void
ba_events_print(void)
{
    ba_print_info("Events %p:\n", events);
    ba_print_info("  id_gen: %p\n", id_gen);
    ba_print_info("  virtual_clock_abs_time: " BA_REAL_FMT "\n",
		  virtual_clock_abs_time);
    g_slist_foreach(events, ba_event_print_from_list, 0);
}

extern ba_status
ba_events_init(void)
{
    events = 0;
    virtual_clock_abs_time = virtual_clock_initial_abs_time;

    id_gen = ba_id_gen_new();

    return BA_STATUS_SUCCESS;
}

extern ba_status
ba_events_destroy(void)
{
    BA_ASSERT(events == 0);

    g_slist_free(events); // kind of moot, since events has to be 0 here.

    ba_id_gen_delete(id_gen);

    return BA_STATUS_SUCCESS;
}

extern void
ba_events_destroy_from_hook(void *unused)
{
    ba_status status;

    status = ba_events_destroy();
    BA_ASSERT(status == BA_STATUS_SUCCESS);
}
