package ba_harness_shared;
use strict;
use vars qw(@ISA @EXPORT @EXPORT_OK %EXPORT_TAGS $VERSION);

use Exporter;
$VERSION = 1.00;
@ISA = qw(Exporter);

@EXPORT = qw( &parse_output &print_output &print_output_stats &test_dir
	      &get_stat_field &write_file %dists $line_sep &get_trace_info
	      &parse_trace &print_trace $ba_harness_ver &ba_harness_ver_get
	      &get_hostname &ba_harness_prog_get &ba_harness_random_seed_get
	      &make_dir &dist_to_list %trace_args @trace_args_list
	      &make_sched_str &write_id_file &get_uniq_from_list &get_val );
@EXPORT_OK = qw( );
%EXPORT_TAGS = ( );

use vars qw( %dists $line_sep $ba_harness_ver $ba_harness_prog
	     $ba_harness_random_seed %trace_args @trace_args_list );

######################################################################

use constant false => 0;
use constant true => 1;
use Sys::Hostname;

my $has_gotten_host;
my $fqdn;

# trace_args in a hash format (to test for existence)
%trace_args = (num_users => true,
	       virtual_clock_terminate => true,
	       max_req_dis_tasks_per_user => true,
	       user_arrivals => true,
	       task_set_change_range => true,
	       tasks_per_task_set => true,
	       service_time => true,
	       think_time => true);

# trace_args in a list format (to iterate through all easily)
# (i really should be making this from %trace_args)
@trace_args_list = ("num_users",
		    "virtual_clock_terminate",
		    "max_req_dis_tasks_per_user",
		    "user_arrivals", "task_set_change_range",
		    "tasks_per_task_set", "service_time", "think_time");

%dists = (poisson => "poisson",
	  exponential => "exponential",
	  pareto => "pareto",
	  uniform => "uniform",
	  bernoulli => "bernoulli");

$line_sep ="-----------------------------------------------" .
    "-----------------------\n"; 

$ba_harness_prog = "ba_sim";
$ba_harness_ver = 0.1;
$ba_harness_random_seed = 1;

$has_gotten_host = false;

######################################################################

sub get_hostname
{
    my ($hostname);

    if($has_gotten_host == false) {
	$has_gotten_host = true;
	$hostname = hostname();
	($fqdn, undef, undef, undef, undef) = gethostbyname($hostname);
    }
    
    if(!defined($fqdn)) {
	$fqdn = "(unknown)";
    }

    return $fqdn;
}

sub ba_harness_prog_get
{
    return $ba_harness_prog;
}

sub ba_harness_ver_get
{
    return $ba_harness_ver;
}

sub ba_harness_random_seed_get
{
    return $ba_harness_random_seed;
}

sub parse_output
{
    my $ba_output = shift;
    my @lines;
    my %parsed_output;
    my $line;
    my @tmp_list;
    my $tmp_line;

    @lines = split /\n/, $ba_output;

    &parse_output_header(\@lines, "---- proem ----", true);
    $parsed_output{program} = &parse_output_str(\@lines, "program");
    $parsed_output{args} = &parse_opt_output_str(\@lines, "args");
    $parsed_output{run_date} = &parse_output_str(\@lines, "run date");
    $parsed_output{random_seed} = &parse_output_str(\@lines, "random seed");

    &parse_output_header(\@lines, "---- scheduler parameters ----", false);
    $parsed_output{warmup_time} =
	&parse_output_vals(\@lines, "warmup time (for selected metrics)", 1);
    $parsed_output{treat_req_dis_diff} =
	&parse_output_bool(\@lines,
			   "treat requested and disclosed tasks differently");
    @tmp_list =
	&parse_output_str_paren(\@lines, "requested scheduler behavior");
    $parsed_output{req_sched} = $tmp_list[0];
    @tmp_list =
	&parse_output_str_paren(\@lines, "disclosed scheduler behavior");
    $parsed_output{dis_sched} = $tmp_list[0];

    &parse_output_header(\@lines, "---- user behavior ----", false);
    $parsed_output{trace_input} = &parse_output_str(\@lines, "trace input");
    $parsed_output{trace_output} = &parse_output_str(\@lines, "trace output");
    if($parsed_output{trace_input} eq "(none)") {
	&parse_trace_shared(\%parsed_output, \@lines);
    }

    &parse_output_header(\@lines, "---- scheduler ----", false);
    $parsed_output{virtual_clock} =
	&parse_output_vals(\@lines, "virtual clock", 1);
    ($parsed_output{all_cand_queue}, $parsed_output{all_cand_queue_weighed}) =
	&parse_output_vals(\@lines, "all candidates queue", 2);
    ($parsed_output{req_cand_queue}, $parsed_output{req_cand_queue_weighed}) =
	&parse_output_vals(\@lines, "requested candidates queue", 2);
    ($parsed_output{dis_cand_queue}, $parsed_output{dis_cand_queue_weighed}) =
	&parse_output_vals(\@lines, "disclosed candidates queue", 2);
       
    &parse_output_header(\@lines, "---- resources ----", false);
    $parsed_output{all_res_usage} = &parse_output_stats(\@lines,
							"all resource usage");
    $parsed_output{req_res_usage} =
	&parse_output_stats(\@lines, "requested resource usage");
    $parsed_output{dis_res_usage} =
	&parse_output_stats(\@lines, "disclosed only rsrc usage");
    $parsed_output{scaled_res_usage} =
	&parse_output_stats(\@lines, "scaled resource usage (all / req.)");
    $parsed_output{billable_resources} =
	&parse_output_stats(\@lines, "billable resources");
    $parsed_output{res_busy_range_time} =
	&parse_output_vals(\@lines, "resource busy range time", 1);
    @tmp_list =
	&parse_output_str_paren(\@lines, "resource busy time");
    $parsed_output{res_busy_time} = $tmp_list[0];
    $parsed_output{res_idle_time} =
	&parse_output_vals([$tmp_list[1]], "resource idle time", 1);    
    $parsed_output{non_work_cons_time} =
	&parse_output_vals(\@lines, "non-work conserving time", 1);
    $parsed_output{load} =
	&parse_output_vals(\@lines, "load", 1);
    $parsed_output{req_load} =
	&parse_output_vals(\@lines, "requested load", 1);
    $parsed_output{req_load_ignoring_idle} =
	&parse_output_vals(\@lines, "requested load ignoring idle time", 1);

    &parse_output_header(\@lines, "---- tasks and task sets ----", false);
    $parsed_output{user_count} =
	&parse_output_vals(\@lines, "user count", 1);
    $parsed_output{num_hit_req_dis_limit} =
	&parse_output_vals(\@lines, "times the req. & dis. max was reached",
			   1);
    $parsed_output{task_sets} = &parse_output_stats(\@lines, "task sets");
    @tmp_list =
	&parse_output_str_paren(\@lines, "task sets alive time");
    $parsed_output{task_sets_alive_time} = $tmp_list[0];
    $parsed_output{no_task_sets_alive_time} =
	&parse_output_vals([$tmp_list[1]], "no task sets alive time", 1);    
    $parsed_output{tasks_ready_time} =
	&parse_output_vals(\@lines, "tasks ready time", 1);
    $parsed_output{disclosed_tasks} = &parse_output_stats(\@lines,
							  "disclosed tasks");
    $parsed_output{disclosed_only_tasks} =
	&parse_output_stats(\@lines, "disclosed only tasks");
    $parsed_output{executed_tasks} = &parse_output_stats(\@lines,
							 "executed tasks");
    $parsed_output{requested_tasks} = &parse_output_stats(\@lines,
							  "requested tasks");
    $parsed_output{finished_tasks} = &parse_output_stats(\@lines,
							 "finished tasks");
    $parsed_output{canceled_tasks} = &parse_output_stats(\@lines,
							 "canceled tasks");

    &parse_output_header(\@lines, "---- visible response time ----", false);
    $parsed_output{vrt} =
	&parse_output_stats(\@lines, "visible response time");
    $parsed_output{vsd} =
	&parse_output_stats(\@lines, "visible slowdown (vrt / st)");
    $parsed_output{user_vrt} =
	&parse_output_stats(\@lines, "u. visible response time");
    $parsed_output{user_vsd} =
	&parse_output_stats(\@lines, "u. visible slowdown");
    $parsed_output{deadlines_met} =
	&parse_output_stats(\@lines, "deadlines met");

    if (@lines != 0) {
	print "@lines";
	die "parse error";
    }

    return \%parsed_output;
}

sub parse_output_header
{
    my $lines = shift;
    my $text = shift;
    my $first = shift;
    my $line;

    if($first == false) {
	$line = shift @$lines;
	die "parse error" if($line ne "");
    }
    $line = shift @$lines;
    die "parse error" if($text ne $line);
}

sub parse_output_str
{
    my $lines = shift;
    my $text = quotemeta shift;
    my $num_items = shift;
    my $line;
    my $str;

    $line = shift @$lines;
    if($line =~ /^$text: (.+)/) {
	$str = $1;
    } else {
	die "parse error";
    }

    return $str;
}

sub parse_opt_output_str
{
    my $lines = shift;
    my $text = quotemeta shift;
    my $num_items = shift;
    my $line;
    my $str;

    $line = shift @$lines;
    if($line =~ /^$text: (.*)/) {
	$str = $1;
    } else {
	die "parse error: `$line'";
    }

    return $str;
}

sub parse_output_bool
{
    my $lines = shift;
    my $text = shift;
    my $bool;

    $bool = &parse_output_str($lines, $text);

    if($bool eq "BA_FALSE") {
	return false;
    } elsif($bool eq "BA_TRUE") {
	return true;
    } else {
	die "parse error";
    }
}

sub parse_output_str_paren
{
    my $lines = shift;
    my $text = quotemeta shift;
    my $num_items = shift;
    my $line;
    my $str;
    my $paren;
    my @result;

    $line = shift @$lines;

    if($line =~ /^$text: ([^(]+) \((.*)\)$/) {
	$str = $1;
	$paren = $2;
    } else {
	die "parse error";
    }

    @result = ($str, $paren);
    return @result;
}

sub parse_output_vals
{
    my $lines = shift;
    my $text = quotemeta shift;
    my $num_items = shift;
    my $line;
    my @items;

    $line = shift @$lines;
    if($line =~ /^$text: (.+)$/) {

	@items = split /, /, $1; # / this slash fixes font-lock
	die "parse error" if @items != $num_items;
    } else {
	die "parse error";
    }

    if(wantarray) {
	return @items;
    } else {
	return $items[0];
    }
}

sub parse_output_dist
{
    my $lines = shift;
    my $text = quotemeta shift;
    my $num_items = shift;
    my $line;
    my $dist;
    my $dist_name;

    $line = shift @$lines;
    if($line =~ /^$text: (.+)/) {
	$dist = $1;
    } else {
	die "parse error";
    }

    if($dist =~ /^(.+):(.+)$/) {
	$dist_name = $1;
    } else {
	die "parse error";
    }

    if(!defined($dists{$dist_name})) {
	die "parse error";
    }

    return $dist;
}

sub parse_output_stats
{
    my $lines = shift;
    my $text = quotemeta shift;
    my $line;
    my @stats;
    my @tmp_stats;
    my $range_str;
    my @range;

    $line = shift @$lines;
    if($line =~ /^$text: (.+)/) {

	@tmp_stats = split /, /, $1; # / this slash fixes font-lock

	if(@tmp_stats != 5) {
	    print STDERR "line: $line\n";
	    die "parse error";
	}
    } else {
	die "parse error";
    }

    $range_str = $tmp_stats[2];
    if($range_str =~ /(-?[^-]+)-(-?[^-]+)/) {
	@range = ($1, $2);
    } else {
	die "parse error";
    }

    @stats = ($tmp_stats[0], $tmp_stats[1], @range, $tmp_stats[3],
	      $tmp_stats[4]);

    return \@stats;
}

sub get_val
{
    my $metric = shift;
    my $field = shift;
    my $val;

    die "no metric?!" if !defined($metric);

    if(defined($field)) {
	$val = &get_stat_field($metric, $field);
    } else {
	$val = $metric;
    }

    die "couldn't find val" if !defined($val);

    return $val;
}

sub get_stat_field
{
    my $stat_ref = shift;
    my $field = shift;
    my @stat;

    @stat = @$stat_ref;

    die "wrong num of fields" if(@stat != 6);

    if($field eq "mean") {
	return $stat[0];
    } elsif($field eq "total") {
	return $stat[1];
    } elsif($field eq "min") {
	return $stat[2];
    } elsif($field eq "max") {
	return $stat[3];
    } elsif($field eq "stddev") {
	return $stat[4];
    } elsif($field eq "count") {
	return $stat[5];
    } else {
	die "wanted `$field' from a stat.";
    }
}

sub get_trace_info
{
    my $trace_file = shift;
    my $line;
    my $str;
    my $found;

    open FILE, "<$trace_file" or return undef;

    $str = "";

    $found = false;
    while(<FILE>) {
	$line = $_;
	if($line =~ /^---- trace ----\n$/) {
	    $found = true;
	    last;
	}
	$str .= $line;
    }
    die "couldn't parse `$trace_file'" if($found == false);

    close FILE or die "can't close file `$trace_file'";

    return $str;
}

sub parse_trace
{
    my $trace_info = shift;
    my @lines;
    my %parsed_trace;

    @lines = split /\n/, $trace_info;

    &parse_output_header(\@lines, "---- proem ----", true);
    $parsed_trace{program} = &parse_output_str(\@lines, "program");
    $parsed_trace{args} = &parse_opt_output_str(\@lines, "args");
    $parsed_trace{run_date} = &parse_output_str(\@lines, "run date");
    $parsed_trace{random_seed} = &parse_output_str(\@lines, "random seed");

    &parse_output_header(\@lines, "---- user behavior ----", false);
    $parsed_trace{trace_input} = &parse_output_str(\@lines, "trace input");
    die if($parsed_trace{trace_input} ne "(none)");
    $parsed_trace{trace_output} = &parse_output_str(\@lines, "trace output");
    die if($parsed_trace{trace_output} eq "(none)");
    &parse_trace_shared(\%parsed_trace, \@lines);

    return \%parsed_trace;
}

sub parse_trace_shared
{
    my ($behavior, $lines) = @_;

    $behavior->{num_users} =
	&parse_output_vals($lines, "number of users", 1);
    $behavior->{virtual_clock_terminate} =
	&parse_output_vals($lines, "virtual clock terminate", 1);
    $behavior->{max_req_dis_tasks_per_user} =
	&parse_output_vals($lines,
			   "maximum requested and disclosed only " .
			   "tasks per user", 1);
    $behavior->{user_arrivals} =
	&parse_output_dist($lines, "user arrivals");
    $behavior->{task_set_change_range} =
	&parse_output_dist($lines, "task set change range");
    $behavior->{tasks_per_task_set} =
	&parse_output_dist($lines, "tasks per task set");
    $behavior->{service_time} = &parse_output_dist($lines, "service time");
    $behavior->{think_time} = &parse_output_dist($lines, "think time");
}

# probably should be using %dists to simplify things
sub dist_to_list
{
    my $dist = shift;
    my ($kind, $alpha, $beta, $mean, $prob);
    my $vals_str;
    my @vals;
    my @ret_vals;

    if($dist =~ /^(.*):(.*)/) {
	$kind = $1;
	$vals_str = $2;
    } else {
	die "couldn't parse dist.";
    }

    @vals = split /,/, $vals_str; # / this slash fixes font-lock

    if($kind eq "uniform") {
	@ret_vals = ( $vals[0], $vals[1], "NULL", "NULL" );
    } elsif($kind eq "exponential") {
	@ret_vals = ( "NULL", "NULL", $vals[0], "NULL" );
    } elsif($kind eq "pareto") {
	@ret_vals = ( $vals[0], $vals[1], $vals[2], "NULL" );
    } elsif($kind eq "bernoulli") {
	@ret_vals = ( "NULL", "NULL", "NULL", $vals[0] );
    } elsif($kind eq "poisson") {
	@ret_vals = ( "NULL", "NULL", $vals[0], "NULL" );
    } else {
	die "unknown dist kind";
    }

    return $kind, @ret_vals;
}

sub make_sched_str
{
    my $output = shift;

    if($output->{dis_sched} eq "NOOP") {
	return "$output->{req_sched}, inter";
    } elsif($output->{dis_sched} eq "n/a") {
	return "$output->{req_sched}, batch";
    } else {
	return "$output->{req_sched} x $output->{dis_sched}, ba";
    }
}

######################################################################

sub get_uniq_from_list
{
    my $list = shift;
    my %helper;
    my @uniq;

    foreach my $elem (@$list) {
	$helper{$elem}++;
    }

    return [keys %helper];
}

sub write_id_file
{
    my ($file, $list) = @_;
    my $uniq_list;
    my $id_str;

    return unless defined $file;

    $uniq_list = &get_uniq_from_list($list);

    $id_str = join("\n", @$uniq_list);
    $id_str .= "\n";
    &write_file($file, $id_str);    
}

######################################################################

sub print_output
{
    my $parsed_output = shift;

    print "program: $$parsed_output{program}\n";
    print "args: $$parsed_output{args}\n";
    print "run date: $$parsed_output{run_date}\n";
    print "random seed: $$parsed_output{random_seed}\n";
    print "treat req dis diff: $$parsed_output{treat_req_dis_diff}\n";
    print "requested scheduler behavior: $$parsed_output{req_sched}\n";
    print "disclosed scheduler behavior: $$parsed_output{dis_sched}\n";
    print "trace input: $$parsed_output{trace_input}\n";
    print "trace output: $$parsed_output{trace_output}\n";
    print "virtual clock: $$parsed_output{virtual_clock}\n";
    print "all cand queue: $$parsed_output{all_cand_queue}\n";
    print "all cand queue weighed: $$parsed_output{all_cand_queue_weighed}\n";
    print "req cand queue: $$parsed_output{req_cand_queue}\n";
    print "req cand queue weighed: $$parsed_output{req_cand_queue_weighed}\n";
    print "dis cand queue: $$parsed_output{dis_cand_queue}\n";
    print "dis cand queue weighed: $$parsed_output{dis_cand_queue_weighed}\n";
    print "all res usage: ";
    &print_output_stats($$parsed_output{all_res_usage}); print "\n";
    print "req res usage: ";
    &print_output_stats($$parsed_output{req_res_usage}); print "\n";
    print "dis res usage: ";
    &print_output_stats($$parsed_output{dis_res_usage}); print "\n";
    print "scaled resource usage: ";
    &print_output_stats($$parsed_output{scaled_res_usage}); print "\n";
    print "billable resources: ";
    &print_output_stats($$parsed_output{billable_resources}); print "\n";
    print "res busy range time: $$parsed_output{res_busy_range_time}\n";
    print "res busy time: $$parsed_output{res_busy_time}\n";
    print "res idle time: $$parsed_output{res_idle_time}\n";
    print "non work cons time: $$parsed_output{non_work_cons_time}\n";
    print "load: $$parsed_output{load}\n";
    print "req load: $$parsed_output{req_load}\n";
    print "req load ignoring idle: $$parsed_output{req_load_ignoring_idle}\n";
    print "user count: $$parsed_output{user_count}\n";
    print "num hit req dis limit: $$parsed_output{num_hit_req_dis_limit}\n";
    print "task sets: ";
    &print_output_stats($$parsed_output{task_sets}); print "\n";
    print "task sets alive time: $$parsed_output{task_sets_alive_time}\n";
    print "no task sets alive time: " .
	"$$parsed_output{no_task_sets_alive_time}\n";
    print "tasks ready time: $$parsed_output{tasks_ready_time}\n";
    print "disclosed tasks: ";
    &print_output_stats($$parsed_output{disclosed_tasks}); print "\n";
    print "disclosed only tasks: ";
    &print_output_stats($$parsed_output{disclosed_only_tasks}); print "\n";
    print "executed tasks: ";
    &print_output_stats($$parsed_output{executed_tasks}); print "\n";
    print "requested tasks: ";
    &print_output_stats($$parsed_output{requested_tasks}); print "\n";
    print "finished tasks: ";
    &print_output_stats($$parsed_output{finished_tasks}); print "\n";
    print "canceled tasks: ";
    &print_output_stats($$parsed_output{canceled_tasks}); print "\n";
    print "vrt: ";
    &print_output_stats($$parsed_output{vrt}); print "\n";
    print "vsd: ";
    &print_output_stats($$parsed_output{vsd}); print "\n";
    print "user vrt: ";
    &print_output_stats($$parsed_output{user_vrt}); print "\n";
    print "user vsd: ";
    &print_output_stats($$parsed_output{user_vsd}); print "\n";
    print "deadlines met: ";
    &print_output_stats($$parsed_output{deadlines_met}); print "\n";
}

sub print_trace
{
    my $parsed_trace = shift;

    print "program: $$parsed_trace{program}\n";
    print "args: $$parsed_trace{args}\n";
    print "run date: $$parsed_trace{run_date}\n";
    print "random seed: $$parsed_trace{random_seed}\n";
    print "trace input: $$parsed_trace{trace_input}\n";
    print "trace output: $$parsed_trace{trace_output}\n";
    print "num users: $$parsed_trace{num_users}\n";
    print "virtual clock terminate: $$parsed_trace{virtual_clock_terminate}\n";
    print "max req dis tasks per user: " .
	"$$parsed_trace{max_req_dis_tasks_per_user}\n";
    print "user arrivals: $$parsed_trace{user_arrivals}\n";
    print "task set change range: " .
	"$$parsed_trace{task_set_change_range}\n";
    print "tasks per task set: $$parsed_trace{tasks_per_task_set}\n";
    print "service time: $$parsed_trace{service_time}\n";
    print "think time: $$parsed_trace{think_time}\n";
}

sub print_output_stats
{
    my $stats = shift;
    my $first;

    $first = true;
    foreach my $stat (@$stats) {
	$first ? $first = false : (print ", ");
	print $stat;
    }
}

######################################################################

sub test_dir
{
    my $dir = shift;
    my $ok;

    if(-r $dir && -w $dir && -x $dir && -d $dir) {
	return true;
    } else {
	return false;
    }
}

sub write_file
{
    my ($name, $output) = @_;

    open FILE, ">$name" or die "can't open file `$name' for writing";
    print FILE $output;
    close FILE or die "can't close file `$name'";
}

sub make_dir
{
    my $dir = shift;
    my $rc;

    $rc = mkdir $dir, 0755;
    die "couldn't create `$dir' ($!)" if($rc == 0);
}

######################################################################

1;
