#!/usr/bin/perl -w

die "Change this script to use a version of Perl greater than or equal to 5."
    if $PERL_VERSION < 5;

use strict;
use English;
use Getopt::Long;

use DBI;
use ba_mysql_shared;

use ba_harness_shared;

use constant false => 0;
use constant true => 1;

#use lib "../../bin";
use ba_stat_nums;

my @ids;
my $id_type;

# should be options:
my $matlab_ratio = 0.75;
my $inverse_golden_ratio = (1 / 1.61803399);
my $font_size = 9;
my $image_length = 3.25;
my $image_height = $image_length * $matlab_ratio;
my $axis_length_offset = 0.75;
my $axis_height_offset = $axis_length_offset * $matlab_ratio;
my $marker_size = 5;

# if id_type == analyze:
my ($analyses, $analyses_id);
my ($baseline_outputs, $baseline_outputs_id);
my ($new_outputs, $new_outputs_id);

# if id_type = output:
my ($outputs, $outputs_id);

# no matter what id_type is:
my ($traces, $traces_id);

###

my $metric_name;
my $metric_field;

# options:
my $options_db_host;
my $options_command;
my $options_comment_char = "%";
my $options_graph = false; # used for some of the commands
my $options_graph_type = "bar";
my $options_metric = undef; # used for some of the commands
my $options_write = undef; # used for some of the commands
my $options_behavior = undef; # used for some of the commands (+num_users)
my $from_opt_behavior_dir;
my $from_opt_behavior;

# dump-analyze opts:

# cumul opts:

# trace-ranges opts:

# compare-ba opts:
my $options_compare_ba_improve = false;

# regress opts:
my $options_regress_use_analyze = false;
my $options_regress_ci = 95;
my $options_regress_degree = 1;
my $options_regress_log = false;
my $options_regress_trans_tscr = false;
# not really options, since i don't expect them to change:
my $regress_regular_prec = "%.6f";
my $regress_log_prec = "%.9f";
my $regress_subplot_p1_height_correct = 0.85;
my $regress_subplot_p2_height_correct = 0.65;
my $regress_dot_marker_str = "'.','MarkerSize',5";

# conf-ints opts:

# improve-stats opts:

# sru-hack opts:

# deadlines opts:

# varvrt opts:

# varurr opts:

my $options_take_req = false;

# sbrcomp opts:

# both-dep opts:
my $options_both_dep_deps = undef;
my $both_dep_closure_hack_x_dir;

# ttst opts:

######################################################################

&parse_options;
&setup;

&present;

exit 0;

sub setup
{
    &setup_hashes;
    &show_settings;
    &set_id_type;
}

sub setup_hashes
{
    $analyses_id = &zero_hash;
    $baseline_outputs_id = &zero_hash;
    $new_outputs_id = &zero_hash;
    $outputs_id = &zero_hash;
    $traces_id = &zero_hash;
}

sub set_id_type
{
    if($options_command eq "dump-analyze") {
	$id_type = "analyze";
    } elsif($options_command eq "cumul") {
	$id_type = "analyze";
	# why not `output' too, for doing cumul of an arbitrary
	# metric, like load?  (low-priority)
    } elsif($options_command eq "compare-ba") {
	$id_type = "analyze";
	# can get the same info from `output', but it's easier to use
	# `analyze' ids, since the improvement calc is in there
    } elsif($options_command eq "trace-ranges") {
	$id_type = "analyze";
	# more natural with `output', but it's already implemented to
	# use `trace-ranges'.  even more natural using `trace' ids,
	# which ba_harness is capable of outputting.  eventually
	# implement to use output and trace ids too, since those may
	# be more readily available.  (low-priority)
    } elsif($options_command eq "regress") {
	if($options_regress_use_analyze) {
	    $id_type = "analyze";
	} else {
	    $id_type = "output";
	}
    } elsif($options_command eq "conf-ints") {
	$id_type = "output";
    } elsif($options_command eq "improve-stats") {
	$id_type = "analyze";
    } elsif($options_command eq "sru-hack") {
	$id_type = "output";
    } elsif($options_command eq "deadlines") {
	$id_type = "output";
    } elsif($options_command eq "varvrt") {
	$id_type = "output";
    } elsif($options_command eq "varurr") {
	$id_type = "output";
    } elsif($options_command eq "sbrcomp") {
	$id_type = "output";
    } elsif($options_command eq "both-dep") {
	$id_type = "output";
    } elsif($options_command eq "ttst") {
	$id_type = "output";
    } elsif($options_command eq "output-metric") {
	$id_type = "output";
    } elsif($options_command eq "uncharged-load") {
	$id_type = "output";
    } else {
	die;
    }

    die "must specify metric when using output ids"
	if !defined($options_metric) && $id_type eq "output" &&
	$options_command ne "both-dep" && $options_command ne "uncharged-load"
	&& $options_command ne "varurr";
    die "can't specify metric when not using output ids"
	if defined($options_metric) && $id_type ne "output";
    &parse_metric;

    print STDERR "id_type: $id_type\n";
}

sub parse_metric
{
    return if(!defined($options_metric));

    if($options_metric =~ /^([^:]+):([^:]+)$/) {
	($metric_name, $metric_field) = ($1, $2);
    } elsif($options_metric =~ /^([^:]+)$/) {
	$metric_name = $1;
	$metric_field = undef;
    } else {
	die "bad metric `$options_metric'";
    }
}

######################################################################

sub present
{
    &db_ba_connect($ba_harness_ver, $options_db_host, "dpetrou",
		   $db_dpetrou_pw);

    &get_ids;
    if($id_type eq "trace") {
	&setup_trace;
    } elsif($id_type eq "output") {
	&setup_output;
    } elsif($id_type eq "analyze") {
	&setup_analyze;
    } else {
	die;
    }

    if($options_command eq "dump-analyze") {
	&dump_analyze;
    } elsif($options_command eq "cumul") {
	&cumul;
    } elsif($options_command eq "trace-ranges") {
	&trace_ranges;
    } elsif($options_command eq "compare-ba") {
	&compare_ba;
    } elsif($options_command eq "regress") {
	&regress;
    } elsif($options_command eq "conf-ints") {
	&conf_ints;
    } elsif($options_command eq "improve-stats") {
	&improve_stats;
    } elsif($options_command eq "sru-hack") {
	&sru_hack;
    } elsif($options_command eq "deadlines") {
	&deadlines;
    } elsif($options_command eq "varvrt") {
	&varvrt;
    } elsif($options_command eq "varurr") {
	&varurr;
    } elsif($options_command eq "sbrcomp") {
	&sbrcomp;
    } elsif($options_command eq "both-dep") {
	&both_dep;
    } elsif($options_command eq "ttst") {
	&ttst;
    } elsif($options_command eq "output-metric") {
	&output_metric;
    } elsif($options_command eq "uncharged-load") {
	&uncharged_load;
    } else {
	die;
    }

    &db_disconnect;
}

sub get_ids
{
    my $test_list;

    while(<STDIN>) {
	chop;
	push @ids, $_;
    }
    close STDIN or die "pain";

    # make sure there are no duplicates
    $test_list = &get_uniq_from_list(\@ids);
    die "there were id duplicates" if(@$test_list != @ids);
}

sub zero_hash
{
    my %zero;

    %zero = ( );
    return \%zero;
}

######################################################################

sub setup_trace
{
    &show_table_stats_for_trace;
    &get_traces_for_trace;
}

sub show_table_stats_for_trace
{
    my ($total_count, $ids_count);

    print STDERR "\n";
    $total_count = &db_select_trace_where_project_count;
    $ids_count = @ids;
    print STDERR "trace rows: $total_count (total), " .
	"$ids_count (specified ids)\n";
}

sub get_traces_for_trace
{
    $traces = &db_select_trace_where_ids("t", undef, \@ids);
    die "how could we not get anything?" if(@$traces == 0);
    &make_list_into_hash_by_id($traces, $traces_id);
}

######################################################################

sub setup_output
{
    &show_table_stats_for_output;
    &get_outputs_for_output;
    &get_traces_for_output;
}

sub show_table_stats_for_output
{
    my ($total_count, $ids_count);

    print STDERR "\n";
    $total_count = &db_select_output_where_project_count;
    $ids_count = @ids;
    print STDERR "output rows: $total_count (total), " .
	"$ids_count (specified ids)\n";
}

sub get_outputs_for_output
{
    $outputs = &db_select_output_where_ids("o", undef, \@ids);
    die "how could we not get anything?" if(@$outputs == 0);
    &make_list_into_hash_by_id($outputs, $outputs_id);
}

sub get_traces_for_output
{
    my $trace_ids;

    $trace_ids = &get_uniq_trace_ids_for_output;
    $traces = &db_select_trace_where_ids("t", undef, $trace_ids);
    die "shouldn't have happened" if @$traces == 0;
    &make_list_into_hash_by_id($traces, $traces_id);
}

sub get_uniq_trace_ids_for_output
{
    my (@trace_ids, $uniq_trace_ids);

    foreach my $output (@$outputs) {
	my $trace_id = $output->{trace_input_id};

	push @trace_ids, $trace_id;
    }

    $uniq_trace_ids = &get_uniq_from_list(\@trace_ids);

    return $uniq_trace_ids;
}

######################################################################

sub setup_analyze
{
    &show_table_stats_for_analyze;
    &get_analyses_for_analyze;
    &get_outputs_and_traces_for_analyze;
}

sub show_table_stats_for_analyze
{
    my ($total_count, $ids_count);

    print STDERR "\n";
    $total_count = &db_select_analyze_where_project_count;
    $ids_count = @ids;
    print STDERR "analyze rows: $total_count (total), " .
	"$ids_count (specified ids)\n";
}

sub get_analyses_for_analyze
{
    $analyses = &db_select_analyze_where_ids("a", undef, \@ids);
    die "how could we not get anything?" if(@$analyses == 0);
    &make_list_into_hash_by_id($analyses, $analyses_id);
}

sub get_outputs_and_traces_for_analyze
{
    my ($trace_ids, $baseline_ids, $new_ids);

    ($trace_ids, $baseline_ids, $new_ids) =
	&get_uniq_trace_and_output_ids_for_analyze;

    $traces = &db_select_trace_where_ids("t", undef, $trace_ids);
    $baseline_outputs = &db_select_output_where_ids("o", undef, $baseline_ids);
    $new_outputs = &db_select_output_where_ids("o", undef, $new_ids);
    die "shouldn't have happened" if @$traces == 0 ||
	@$baseline_outputs == 0 || @$new_outputs == 0;

    &make_list_into_hash_by_id($traces, $traces_id);
    &make_list_into_hash_by_id($baseline_outputs, $baseline_outputs_id);
    &make_list_into_hash_by_id($new_outputs, $new_outputs_id);
}

sub get_uniq_trace_and_output_ids_for_analyze
{
    my (@trace_ids, $uniq_trace_ids);
    my (@baseline_ids, $uniq_baseline_ids);
    my (@new_ids, $uniq_new_ids);

    foreach my $analyze (@$analyses) {
	my $trace_id = $analyze->{trace_input_id};
	my $baseline_id = $analyze->{baseline_id};
	my $new_id = $analyze->{new_id};

	push @trace_ids, $trace_id;
	push @baseline_ids, $baseline_id;
	push @new_ids, $new_id;
    }

    $uniq_trace_ids = &get_uniq_from_list(\@trace_ids);
    $uniq_baseline_ids = &get_uniq_from_list(\@baseline_ids);
    $uniq_new_ids = &get_uniq_from_list(\@new_ids);

    return ($uniq_trace_ids, $uniq_baseline_ids, $uniq_new_ids);
}

######################################################################

sub dump_analyze
{
    print "id\ttrace_input_id\tmetric_order\tmetric_name\tmetric_field\t" .
	"baseline_id\tbaseline_val\tnew_id\tnew_val\timprovement\n";
    foreach my $analyze (@$analyses) {
	my ($id, $trace_input_id, $metric_order, $metric_name, $metric_field,
	    $baseline_id, $baseline_val, $new_id, $new_val, $improvement);

	$id = $analyze->{id};
	$trace_input_id = $analyze->{trace_input_id};
	$metric_order = $analyze->{metric_order};
	$metric_name = $analyze->{metric_name};
	if(defined($analyze->{metric_field})) {
	    $metric_field = $analyze->{metric_field};
	} else {
	    $metric_field = "(none)";
	}
	$baseline_id = $analyze->{baseline_id};
	$baseline_val = $analyze->{baseline_val};
	$new_id = $analyze->{new_id};
	$new_val = $analyze->{new_val};
	$improvement = $analyze->{improvement};

	print "$id\t$trace_input_id\t$metric_order\t$metric_name\t" .
	    "$metric_field\t$baseline_id\t$baseline_val\t$new_id\t" .
	    "$new_val\t$improvement\n";
    }
}

######################################################################

sub cumul
{
    my @improvements;
    my $count;
    my @cumul;
    my ($baseline_sched_type, $new_sched_type);

    ($baseline_sched_type, $new_sched_type) = &cumul_checks;

    &cumul_show_header($baseline_sched_type, $new_sched_type);

    # actual computation start
    foreach my $analyze (@$analyses) {
	my $improvement;

	$improvement = $analyze->{improvement};
	push @improvements, $improvement;
    }

    @improvements = sort sort_numerically_callback @improvements;

    $count = @ids;    
    for(my $i = 0; $i < $count; $i++) {
	my $val = $improvements[$i];
	my $frac = ($count - $i) / $count;
	my $nice_frac;

	$nice_frac = sprintf "%.3f", $frac;

	push @cumul, [$val, $nice_frac];
    }
    # actual computation end

    if($options_graph) {
	&cumul_show_graph(\@cumul);
    } else {
	&cumul_show_non_graph(\@cumul);
    }
}

sub cumul_show_header
{
    my ($baseline_sched_type, $new_sched_type) = @_;
    my @header_lines;
    my $metric_str = &make_metric_str(true);

    push @header_lines, "metric: $metric_str\n";
    push @header_lines, &cumul_make_sched_header_lines($baseline_sched_type,
						       $new_sched_type);
    push @header_lines, &show_header_trace_range_helper;

    # send lines to stdout
    foreach my $line (@header_lines) {
	print "$options_comment_char $line";
    }
    print "\n";

    if($options_graph) {
	# send munged lines to stdout that will tell matlab to display
	# lines.
	foreach my $line (@header_lines) {
	    chop $line;
	    print "disp('$line')\n";
	}
	print "\n";
    }
}

sub cumul_make_sched_header_lines
{
    my ($baseline_sched_type, $new_sched_type) = @_;
    my @lines;

    push @lines, "baseline sched: $baseline_sched_type, " .
	"new sched: $new_sched_type\n";
    push @lines, "\n";

    return @lines;
}

sub cumul_checks
{
    my $first;
    my ($first_baseline_sched_type, $first_new_sched_type);

    # ensure only one kind of baseline (inter/ba/batch) vs. new
    # (inter/ba/batch).
    $first = true;
    foreach my $analyze (@$analyses) {
	my ($baseline_id, $new_id);
	my ($baseline_output, $new_output);

	$baseline_id = $analyze->{baseline_id};
	$new_id = $analyze->{new_id};
	$baseline_output = $baseline_outputs_id->{$baseline_id};
	$new_output = $new_outputs_id->{$new_id};

	if($first) {
	    $first = false;
	    $first_baseline_sched_type =
		&get_sched_type($baseline_output);
	    $first_new_sched_type = &get_sched_type($new_output);
	    next;
	}

	die "sched type for baseline not the same in cumul"
	    if($first_baseline_sched_type ne
	       &get_sched_type($baseline_output));
	die "sched type for new not the same in cumul"
	    if($first_new_sched_type ne &get_sched_type($new_output));
    }

    # ensure there is only one type of metric
    (undef, $metric_name, $metric_field) =
	&check_one_type_of_metric_for_analyze;

    return ($first_baseline_sched_type, $first_new_sched_type);
}

sub cumul_show_graph
{
    my $cumul = shift;
    my ($xmin, $xmax, $ymin, $ymax);
    my ($first, $last);
    my $length;
    my $metric_str = &make_metric_str(true);

    &show_new_figure;

    &cumul_show_graph_vec_helper($cumul, "x", 0);
    &cumul_show_graph_vec_helper($cumul, "y", 1);

    print("plot(cumul_x, cumul_y)\n");

    $length = length @$cumul;
    $first = $$cumul[0];
    $last = $$cumul[@$cumul - 1];

    $xmin = $$first[0];
    $xmax = $$last[0];
    $ymin = 0;
    $ymax = 1;

    print "axis([$xmin $xmax $ymin $ymax])\n";
    print "xlabel('improvement')\n";
    print "ylabel('fraction')\n";
    &show_box_off;
}

sub cumul_show_graph_vec_helper
{
    my ($cumul, $str, $ind) = @_;
    my $first;

    print "cumul_$str = [";
    $first = true;
    foreach my $items (@$cumul) {
	if($first) {
	    $first = false;
	} else {
	    print ",";
	}
	print "$$items[$ind]";
    }
    print "];\n";
}

sub cumul_show_non_graph
{
    my $cumul = shift;

    foreach my $items (@$cumul) {
	print "$$items[0] $$items[1]\n";
    }
}

######################################################################

sub trace_ranges
{
    my $lines;

    $lines = &trace_ranges_helper;
    foreach my $line (@$lines) {
	print $line;
    }
}

sub trace_ranges_helper
{
    my @lines;

    foreach my $trace_arg (@trace_args_list) {
	push @lines, &trace_ranges_one_arg($trace_arg);
    }

    return \@lines;
}

sub trace_ranges_one_arg
{
    my $trace_arg = shift;
    my ($test_trace, $any_id);
    my @ids;
    my $line;

    $test_trace = $$traces[0];
    if(defined($test_trace->{"${trace_arg}_kind"})) {
	$line = &trace_ranges_one_dist_arg($trace_arg);
    } else {
	$line = &trace_ranges_one_val_arg($trace_arg);
    }

    return $line;
}

sub trace_ranges_one_dist_arg
{
    my $trace_arg = shift;
    my $first;
    my (@min_dist, @max_dist);
    my ($min_str, $max_str);

    $first = true;
    foreach my $trace (@$traces) {
	my @dist = &dist_to_list($trace->{$trace_arg});

	if($first) {
	    $first = false;
	    @min_dist = @max_dist = @dist;
	    next;
	}
	
	die "doing trace-ranges on traces w/ trace args of different dist " .
	    "types" if $dist[0] ne $min_dist[0];
	
	for(my $i = 0; $i < @dist; $i++) {
	    next if $i == 0; # skip kind
	    next if($dist[$i] eq "NULL"); # undef alpha / beta, whatever
	    
	    $min_dist[$i] = $dist[$i] if($dist[$i] < $min_dist[$i]);
	    $max_dist[$i] = $dist[$i] if($dist[$i] > $max_dist[$i]);
	}
    }

    $min_str = &trace_range_dist_helper_get(@min_dist);
    $max_str = &trace_range_dist_helper_get(@max_dist);

    return "$trace_arg: $min_str-$max_str\n";
}

sub trace_range_dist_helper_get
{
    my @dist = @_;
    my $str;
    my $val_str;
    my @vals;

    $str = "$dist[0]:";
    shift @dist;
    foreach my $val (@dist) {
	next if($val eq "NULL");
	push @vals, $val;
    }
    $val_str = join(",", @vals);
    $str .= "$val_str";

    return $str;
}

sub trace_ranges_one_val_arg
{
    my $trace_arg = shift;
    my ($min_val, $max_val);
    my $first;

    $first = true;
    foreach my $trace (@$traces) {
	my $val;
	
	$val = $trace->{$trace_arg};
	
	if($first) {
	    $first = false;
	    $min_val = $max_val = $val;
	    next;
	}
	
	$min_val = $val if($val < $min_val);
	$max_val = $val if($val > $max_val);
    }

    return "$trace_arg: $min_val-$max_val\n";
}

######################################################################

sub compare_ba
{
    my $set;
    my ($inter_sched, $ba_sched, $batch_sched);

    &set_behavior;

    $set = &db_select_analyze_for_compare_ba(\@ids);
    die "no results!" if @$set == 0;

    ($inter_sched, $ba_sched, $batch_sched) = &compare_ba_checks($set);
    $set = [sort compare_ba_sort_callback @$set];

    &compare_ba_show_header($inter_sched, $ba_sched, $batch_sched);

    if($options_graph) {
	&compare_ba_show_graph($set, $inter_sched, $ba_sched, $batch_sched);
    } else {
	&compare_ba_show_non_graph($set, $inter_sched, $ba_sched, $batch_sched);
    }
}

sub compare_ba_checks
{
    my $set = shift;
    my ($inter_sched, $ba_sched, $batch_sched);

    # ensure there are the same number of sets and traces

    die "num set (" . @$set . ") != num traces (" . @$traces . ")"
	if(@$set != @$traces);

    # ensure all dists have the same kind across traces and ensure
    # that only the specified trace behavior changes
    &check_dist_kind_and_behavior_changes;

    # ensure only ba in new, and only either batch/inter in baseline.
    &compare_ba_checks_ensure_sched_type;

    # and ensure that the scheds in inter/ba/batch never change
    ($inter_sched, $ba_sched, $batch_sched) =
	&compare_ba_checks_ensure_same_scheds($set);

    # ensure there is only one type of metric
    (undef, $metric_name, $metric_field) =
	&check_one_type_of_metric_for_analyze;

    return ($inter_sched, $ba_sched, $batch_sched);
}

sub compare_ba_checks_ensure_sched_type
{
    foreach my $analyze (@$analyses) {
	my ($baseline_id, $new_id);
	my ($baseline_output, $new_output);
	my ($baseline_sched_type, $new_sched_type);

	$baseline_id = $analyze->{baseline_id};
	$new_id = $analyze->{new_id};
	$baseline_output = $baseline_outputs_id->{$baseline_id};
	$new_output = $new_outputs_id->{$new_id};

	$baseline_sched_type = &get_sched_type($baseline_output);
	$new_sched_type = &get_sched_type($new_output);

	die "baseline must be inter or batch for compare-ba"
	    if $baseline_sched_type ne "interactive" &&
	    $baseline_sched_type ne "batch";
	die "new must be ba for compare-ba"
	    if $new_sched_type ne "batchactive";
    }
}

sub compare_ba_checks_ensure_same_scheds
{
    my $set = shift;
    my $first;
    my ($first_inter_sched, $first_ba_sched, $first_batch_sched);

    $first = true;
    foreach my $row (@$set) {
	my ($inter_id, $inter_val, $improvement_over_inter,
	    $ba_id, $ba_val,
	    $batch_id, $batch_val, $improvement_over_batch,
	    $trace_input_id) = @$row;
	my ($inter_sched, $ba_sched, $batch_sched);

	$inter_sched = &make_sched_str($baseline_outputs_id->{$inter_id});
	$ba_sched = &make_sched_str($new_outputs_id->{$ba_id});
	$batch_sched = &make_sched_str($baseline_outputs_id->{$batch_id});

	if($first) {
	    $first = false;
	    $first_inter_sched = $inter_sched;
	    $first_ba_sched = $ba_sched;
	    $first_batch_sched = $batch_sched;
	    next;
	}

	die "inter sched changed" if($inter_sched ne $first_inter_sched);
	die "ba sched changed ($ba_sched != $first_ba_sched)"
	    if($ba_sched ne $first_ba_sched);
	die "batch sched changed" if($batch_sched ne $first_batch_sched);

    }

    return ($first_inter_sched, $first_ba_sched, $first_batch_sched);
}

sub compare_ba_show_header
{
    my ($inter_sched, $ba_sched, $batch_sched) = @_;
    my @header_lines;

    @header_lines = &compare_ba_make_header($inter_sched, $ba_sched,
					    $batch_sched);
    # send lines to stdout
    foreach my $line (@header_lines) {
	print "$options_comment_char $line";
    }
    print "\n";

    if($options_graph) {
	# add some stuff b4hand:
	if($options_compare_ba_improve) {
	    print "disp('compare-ba-improve: true')\n";
	} else {
	    print "disp('compare-ba-improve: false')\n";
	}
	print "disp('behavior: $options_behavior')\n";
	# send munged lines to stdout that will tell matlab to display
	# lines.
	foreach my $line (@header_lines) {
	    chop $line;
	    print "disp('$line')\n";
	}
	print "\n";
    }
}

sub compare_ba_make_header
{
    my ($inter_sched, $ba_sched, $batch_sched) = @_;
    my @lines;
    my $metric_str = &make_metric_str($options_compare_ba_improve);

    push @lines, "metric: $metric_str\n";
    push @lines, "\"$ba_sched\" is compared against \"$inter_sched\" and " .
	"\"$batch_sched\"\n";
    push @lines, &show_header_trace_range_helper;

    return @lines;
}

sub compare_ba_show_graph
{
    my ($set, $inter_sched, $ba_sched, $batch_sched) = @_;
    my $kind;
    my $list_of_trace_ids;

    foreach my $row (@$set) {
	my (undef, undef, undef,
	    undef, undef,
	    undef, undef, undef,
	    $trace_input_id) = @$row;

	push @$list_of_trace_ids, $trace_input_id;
    }
    
    &show_new_figure;
    $kind = &generic_show_graph_x($list_of_trace_ids);

    if($options_graph_type eq "bar") {
	&compare_ba_show_graph_Y($set);
	print "bar(x,Y);\n";
    } elsif($options_graph_type eq "line") {
	my $graph_str;

	&compare_ba_show_graph_multiple_y($set);
	if($options_compare_ba_improve) {
	    $graph_str = join(",",
			      "x", "improvement_over_inter", "'-+k'",
			      "x", "improvement_over_batch", "'--ok'");
	} else {
	    $graph_str = join(",",
			      "x", "batch_val", "'--ok'",
			      "x", "inter_val", "'-+k'",
			      "x", "ba_val", "'-.xk'");
	}
    
	print "plot($graph_str);\n";
    } else {
	die;
    }

    &compare_ba_show_graph_labels($kind);
    &compare_ba_show_graph_legend($inter_sched, $ba_sched, $batch_sched);
    print "legend('boxoff');\n";
    print "grid on;\n";
    print "box off;\n";
    &generic_show_graph_write;
}

sub compare_ba_show_graph_Y
{
    my $set = shift;
    my $first;

    print "Y = [";
    $first = true;
    foreach my $row (@$set) {
	my (undef, $inter_val, $improvement_over_inter,
	    undef, $ba_val,
	    undef, $batch_val, $improvement_over_batch,
	    undef) = @$row;

	if($first) {
	    $first = false;
	} else {
	    print "\n";
	}

	if($options_compare_ba_improve) {
	    print "$improvement_over_inter, $improvement_over_batch";
	} else {
	    print "$batch_val, $inter_val, $ba_val";
	}
    }
    print "];\n";
}

sub compare_ba_show_graph_multiple_y
{
    my $set = shift;

    if($options_compare_ba_improve) {
	&compare_ba_show_graph_single_y($set, "improvement_over_inter");
	&compare_ba_show_graph_single_y($set, "improvement_over_batch");
    } else {
	&compare_ba_show_graph_single_y($set, "batch_val");
	&compare_ba_show_graph_single_y($set, "inter_val");
	&compare_ba_show_graph_single_y($set, "ba_val");
    }
}

sub compare_ba_show_graph_single_y
{
    my ($set, $name) = @_;
    my $first;

    print "$name = [";
    $first = true;
    foreach my $row (@$set) {
	my (undef, $inter_val, $improvement_over_inter,
	    undef, $ba_val,
	    undef, $batch_val, $improvement_over_batch,
	    undef) = @$row;

	if($first) {
	    $first = false;
	} else {
	    print ",";
	}

	print $improvement_over_inter if($name eq "improvement_over_inter");
	print $improvement_over_batch if($name eq "improvement_over_batch");

	print $inter_val if($name eq "inter_val");
	print $ba_val if($name eq "ba_val");
	print $batch_val if($name eq "batch_val");
    }
    print "];\n";
}

sub compare_ba_show_graph_legend
{
    my ($inter_sched, $ba_sched, $batch_sched) = @_;
    my $legend_str;
    my @constant_behaves;
    my $trace;

    $inter_sched = &compare_ba_show_graph_legend_nice_sched($inter_sched);
    $ba_sched = &compare_ba_show_graph_legend_nice_sched($ba_sched);
    $batch_sched = &compare_ba_show_graph_legend_nice_sched($batch_sched);

    $trace = $$traces[0];

    foreach my $trace_arg (@trace_args_list) {
	my $nice_trace_arg;

	next if $trace_arg eq $from_opt_behavior;

	next if $trace_arg eq "num_users";
	next if $trace_arg eq "max_req_dis_tasks_per_user";

	$nice_trace_arg = $trace_arg;
	$nice_trace_arg =~ s/_/\\_/g;

	if(defined($trace->{"${trace_arg}_kind"})) {
	    my @dist = &dist_to_list($trace->{$trace_arg});
	    my $val_str;

	    $val_str = &compare_ba_show_graph_legend_dist_helper(@dist);
	    push @constant_behaves, "'$nice_trace_arg: $val_str'"; # fontcolor
	} else {
	    push @constant_behaves, "'$nice_trace_arg: $trace->{$trace_arg}'";
	}
    }

    $legend_str = join(",", "'$batch_sched'", "'$inter_sched'", "'$ba_sched'");
#    $legend_str = join(",", "'$inter_sched'", "'$ba_sched'", "'$batch_sched'",
#		       @constant_behaves);

    print "legend($legend_str)\n";
    print "colormap('prism');\n";
#    print "colormap('jet');\n";
#    print "colormap('cool');\n";
}

sub compare_ba_show_graph_legend_dist_helper
{
    my @dist = @_;
    my $str;
    my $val_str;
    my @vals;
    my $kind;
    
    $kind = $dist[0];
    shift @dist;

    if($kind eq "exponential") {
	$kind = "exp.";
    } elsif($kind eq "pareto") {
	$kind = "par.";
    } elsif($kind eq "uniform") {
	$kind = "uni.";
    } elsif($kind eq "bernoulli") {
	$kind = "bern.";
    } elsif($kind eq "poisson") {
	$kind = "poi.";
    } else {
	die;
    }

    foreach my $val (@dist) {
	my $nice_val;
	
	next if($val eq "NULL");

	$nice_val = sprintf "%.3f", $val;

	push @vals, $nice_val;
    }
    $val_str = join("-", @vals);
    $str = "$val_str ($kind)";

    return $str;
}

sub compare_ba_show_graph_legend_nice_sched
{
    my $sched = shift;
    my $actual_sched;
    my $sched_type;
    my $ret;

    if($sched =~ /^([^,]+), (.+)$/) {
	$actual_sched = $1;
	$sched_type = $2;
    } else {
	die "couldn't parse sched";
    }

    if($sched_type eq "inter") {
	$sched_type = "inter.";

	$ret = "$actual_sched, $sched_type";
    } elsif($sched_type eq "batch") {
	$ret = "$actual_sched, $sched_type";
    } elsif($sched_type eq "ba") {
	$ret = "$actual_sched";
    } else {
	die;
    }

#    return "$actual_sched ($sched_type)";
    return $ret;
}

sub compare_ba_show_graph_labels
{
    my $kind = shift;
    
    &generic_show_graph_labels($kind, $options_compare_ba_improve);
}

sub compare_ba_show_non_graph
{
    my ($set, $inter_sched, $ba_sched, $batch_sched) = @_;
    my $first = true;

    foreach my $row (@$set) {
	my ($inter_id, $inter_val, $improvement_over_inter,
	    $ba_id, $ba_val,
	    $batch_id, $batch_val, $improvement_over_batch,
	    $trace_input_id) = @$row;
	my $trace;
	
	$first ? $first = false : print "\n";

	$trace = $traces_id->{$trace_input_id};

	print "$options_comment_char $from_opt_behavior " .
	    $trace->{$from_opt_behavior} . "\n";

	if($options_compare_ba_improve) {
	    print "$improvement_over_inter $options_comment_char " .
		"$inter_sched\n";
	    print "$options_comment_char 1.0, of course; $ba_sched\n";
	    print "$improvement_over_batch $options_comment_char" .
		"$batch_sched\n";
	} else {
	    print "$inter_val $options_comment_char $inter_sched\n";
	    print "$ba_val $options_comment_char $ba_sched\n";
	    print "$batch_val $options_comment_char $batch_sched\n";
	}
    }
}

sub compare_ba_sort_callback
{
    my ($set1, $set2);
    my ($set1_trace_input_id, $set2_trace_input_id);

    $set1 = $a; $set2 = $b;

    (undef, undef, undef, undef, undef, undef, undef, undef,
     $set1_trace_input_id) = @$set1;
    (undef, undef, undef, undef, undef, undef, undef, undef,
     $set2_trace_input_id) = @$set2;

    return &sort_on_changing_behavior_helper($set1_trace_input_id,
					     $set2_trace_input_id);
}

######################################################################

sub regress
{
    my ($y_size, $X_num_rows, $X_num_cols, $alpha);
    my ($predictors, $observations);
    my ($sched_a, $sched_b, $trace_refs, $using_trace_args);
    my $num_observations;

    ($sched_a, $sched_b, $using_trace_args) = &regress_checks;

    $alpha = &regress_make_alpha;
    ($observations, $trace_refs) = &regress_make_observations;
    $num_observations = @$observations;
    $predictors = &regress_make_predictors($trace_refs, $using_trace_args);

    &regress_diary_start;

    &regress_show_header($sched_a, $sched_b, $using_trace_args,
			 scalar @$observations);

    $y_size = &regress_show_y_str($observations);
    ($X_num_rows, $X_num_cols) = &regress_show_X_str($predictors);
#    $y_size = &regress_show_y_fake_str($observations);
#    ($X_num_rows, $X_num_cols) = &regress_show_X_fake_str($predictors);
    &regress_checks_dimen($y_size, $X_num_rows, $X_num_cols);

    &regress_show_alpha_str($alpha);

    print "[B, BINT, R, RINT, STATS] = regress(y, X, ALPHA);\n";

    print "B'\n";
    print "BINT\n";
    print "STATS\n";

    &regress_show_graphs($num_observations, $using_trace_args);

    # we want this at the end, since if there's an error, we want at
    # least some of the graphs to be output.
    &regress_show_num_check;

    &regress_diary_end;
}

##########

# checks

sub regress_checks
{
    my ($a, $b, $using_trace_args);

    # make sure that each set of outputs refers to one scheduler
    ($a, $b) = &regress_checks_same_sched;

    # make sure that each output refers to one particular trace
    &regress_checks_single_trace;

    # ensure good dist kinds and eliminates unchanging trace args
    $using_trace_args = &regress_checks_trace_args_iter_and_check;

    if($id_type eq "analyze") {
	# ensure there is only one type of metric
	(undef, $metric_name, $metric_field) =
	    &check_one_type_of_metric_for_analyze;
    }

    return ($a, $b, $using_trace_args);
}

sub regress_checks_same_sched
{
    my ($a, $b);

    if($id_type eq "output") {
	$a = &check_same_sched($outputs);
    } elsif($id_type eq "analyze") {
	$a = &check_same_sched($baseline_outputs);
	$b = &check_same_sched($new_outputs);
    } else {
	die;
    }

    return ($a, $b);
}

sub regress_checks_single_trace
{
    my $size;

    if($id_type eq "output") {
	$size = @$outputs;
    } elsif($id_type eq "analyze") {
	$size = @$analyses;
    } else {
	die;
    }

    die "in regress, two or more outputs referred to the same trace"
	if($size != @$traces);
}

sub regress_checks_trace_args_iter_and_check
{
    my @using_trace_args;

    foreach my $trace_arg (@trace_args_list) {
	my $test_trace;

	$test_trace = $$traces[0];
	if(defined($test_trace->{"${trace_arg}_kind"})) {
	    my ($good_a, $good_b);
	    
	    # eliminate unchanging trace args
	    ($good_a, $good_b) =
		&regress_checks_eliminate_unchanging_dist_helper($trace_arg);
	    push @using_trace_args, [$good_a, $good_b];
	    
	    # ensure good dist kinds
	    &check_good_dist_kind($trace_arg);
	} else {
	    my $good;
	    
	    # eliminate unchanging trace args
	    $good =
		&regress_checks_eliminate_unchanging_val_helper($trace_arg);
	    push @using_trace_args, [$good];
	}
    }

    return \@using_trace_args;
}

sub regress_checks_eliminate_unchanging_val_helper
{
    my $trace_arg = shift;
    my $first;
    my $first_val;
    
    $first = true;
    foreach my $trace (@$traces) {
	my $val = $trace->{$trace_arg};
	
	if($first) {
	    $first = false;
	    $first_val = $val;
	    next;
	}
	
	if($first_val != $val) {
	    return true;
	}
    }
    
    return false;
}

sub regress_checks_eliminate_unchanging_dist_helper
{
    my $trace_arg = shift;
    my $first;
    my ($first_val1, $first_val2);
    
    $first = true;
    
    foreach my $trace (@$traces) {
	my ($val1, $val2) = &get_dist_vals($trace, $trace_arg);
	my ($good1, $good2) = (false, false);
	
	if($first) {
	    $first = false;
	    $first_val1 = $val1;
	    $first_val2 = $val2 if(defined($val2));
	    next;
	}
	
	$good1 = true if($val1 != $first_val1);
	if(defined($first_val2)) {
	    $good2 = true if($val2 != $first_val2);
	    return ($good1, $good2) if($good1 || $good2);
	} else {
	    return ($good1, undef) if($good1);
	}
    }
    if(defined($first_val2)) {
	return (false, false);
    } else {
	return (false, undef);
    }
}

# this occurs later on
sub regress_checks_dimen
{
    my ($y_size, $X_num_rows, $X_num_cols) = @_;

    die "can't have 0 elems (y_size)" if $y_size == 0;
    die "can't have 0 elems (X_num_row)" if $X_num_rows == 0;
    die "num X rows must equal size of y" if($y_size != $X_num_rows);

    # i _think_ the following is valid; may be off by one or something
    die "num cols of X ($X_num_cols) must be less than size of y ($y_size)"
	if($X_num_cols >= $y_size);
}

##########

# header

sub regress_diary_start
{
    my $diary_file;

    return if(!$options_write);

    $diary_file = $options_write . ".txt";
    print "delete $diary_file;\n";
    print "diary $diary_file;\n";
}

sub regress_diary_end
{
    return if(!$options_write);
    print "diary off;\n";
}

sub regress_show_header
{
    my ($sched_a, $sched_b, $using_trace_args, $num_obs) = @_;
    my @lines;
    my $metric_str = &make_metric_str($options_regress_use_analyze);

    if($id_type eq "analyze") {
	push @lines, "metric: $metric_str";
    }
    push @lines, &regress_show_header_sched_helper($sched_a, $sched_b);
    push @lines, &show_header_trace_range_helper;
    push @lines, "\n";
    push @lines,
        &regress_show_header_using_trace_args_helper($using_trace_args,
						     $num_obs);

    # send lines to stdout
    foreach my $line (@lines) {
	print "$options_comment_char $line";
    }
    print "\n";

    # send munged lines to stdout that will tell matlab to display
    # lines.  and i'm prepending some additional information.
    print "disp('metric: $metric_str')\n";
    print "disp('conf. interval: $options_regress_ci%')\n";
    print "disp('poly degree: $options_regress_degree')\n";
    if($options_regress_log) {
	print "disp('take log of inputs/outputs: true')\n";
    } else {
	print "disp('take log of inputs/outputs: false')\n";
    }
    foreach my $line (@lines) {
	chop $line;
	print "disp('$line')\n";
    }
    print "\n";

    print "format compact;\n";
    print "format long g;\n";
}

sub regress_show_header_sched_helper
{
    my ($sched_a, $sched_b) = @_;
    my @lines;

    if($id_type eq "output") {
	push @lines, "scheduler: $sched_a\n";
    } elsif($id_type eq "analyze") {
	push @lines, "baseline scheduler: $sched_a\n";
	push @lines, "new scheduler: $sched_b\n";
    } else {
	die;
    }
    push @lines, "\n";

    return @lines;
}

sub regress_show_header_using_trace_args_helper
{
    my ($using_trace_args, $num_obs) = @_;
    my ($used_names, $skipped_names);
    my ($num_used_names, $num_skipped_names);
    my ($used_names_str, $skipped_names_str);
    my ($num_coeffs, $poly_str, $poly);
    my @lines;
    
    ($used_names, $skipped_names) =
	&regress_make_used_and_skipped_names($using_trace_args);

    $used_names_str = join ", " => @$used_names;
    $skipped_names_str = join ", " => @$skipped_names;
    $num_used_names = @$used_names;
    $num_skipped_names = @$skipped_names;

    push @lines, "used trace args ($num_used_names): $used_names_str\n";
    push @lines, "skipped trace args ($num_skipped_names): " .
	"$skipped_names_str\n";
    push @lines, "number of observations: $num_obs\n";

    ($poly, $num_coeffs) = &regress_make_poly_list_of_used_names($used_names,
								 true);
    $poly_str = join " + ", @$poly;

    push @lines, "solving for the following " . $num_coeffs .
	" constants in: $poly_str\n";

    ($poly, undef) = &regress_make_poly_list_of_used_names($used_names, false);
    push @lines, "param terms list without coefficients:\n";
    my $i = 1;
    foreach my $term (@$poly) {
	push @lines, "    $i: $term\n";
	$i++;
    }

    return @lines;
}

sub regress_make_used_and_skipped_names
{
    my ($using_trace_args_orig, $num_obs) = @_;
    my (@used_names, @skipped_names);
    my @using_trace_args = @$using_trace_args_orig;

    foreach my $trace_arg (@trace_args_list) {
	my $test_trace;
	my $using_trace = shift @using_trace_args;
	
	$test_trace = $$traces[0];
	if(defined($test_trace->{"${trace_arg}_kind"})) {
	    my $kind = $test_trace->{"${trace_arg}_kind"};
	    if($kind eq "exponential") {
		push @used_names, "${trace_arg}_mean" if($$using_trace[0]);
		push @skipped_names, "${trace_arg}_mean" if(!$$using_trace[0]);
	    } elsif($kind eq "uniform") {

		if($options_regress_trans_tscr) {
		    push @used_names, "${trace_arg}_avg" if($$using_trace[0]);
		    push @skipped_names, "${trace_arg}_avg"
			if(!$$using_trace[0]);
		    push @used_names, "${trace_arg}_distance"
			if($$using_trace[1]);
		    push @skipped_names, "${trace_arg}_distance"
			if(!$$using_trace[1]);
		} else {
		    push @used_names, "${trace_arg}_lower"
			if($$using_trace[0]);
		    push @skipped_names, "${trace_arg}_lower"
			if(!$$using_trace[0]);
		    push @used_names, "${trace_arg}_upper"
			if($$using_trace[1]);
		    push @skipped_names, "${trace_arg}_upper"
			if(!$$using_trace[1]);
		}

	    } elsif($kind eq "bernoulli") {
		push @used_names, "${trace_arg}_prob" if($$using_trace[0]);
		push @skipped_names, "${trace_arg}_prob" if(!$$using_trace[0]);
	    } elsif($kind eq "poisson") {
		push @used_names, "${trace_arg}_mean" if($$using_trace[0]);
		push @skipped_names, "${trace_arg}_mean" if(!$$using_trace[0]);
	    } else {
		die;
	    }
	} else {
	    push @used_names, $trace_arg if($$using_trace[0]);
	    push @skipped_names, $trace_arg if(!$$using_trace[0]);
	}
    }

    return (\@used_names, \@skipped_names);
}

sub regress_make_poly_list_of_used_names
{
    my ($used_names, $use_coeff) = @_;
    my @poly;
    my $const_i;

    $const_i = 1;
    if($use_coeff) {
	push @poly, "c_$const_i";
    } else {
	push @poly, "1";
    }
    $const_i++;
    foreach my $used_name (@$used_names) {
	for(my $i = 0; $i < $options_regress_degree; $i++) {
	    my $power;

	    $power = "^" . ($i + 1);
	    $power = "" if($power eq "^1");

	    if($use_coeff) {
		push @poly, "c_$const_i $used_name$power";
	    } else {
		push @poly, "$used_name$power";
	    }
	    $const_i++;
	}
    }

    return (\@poly, $const_i - 1);
}

##########

# make args to matlab's regress

sub regress_make_alpha
{
    return 1 - ($options_regress_ci / 100);
}

sub regress_make_observations
{
    my @observations;
    my @trace_refs;
    
    if($id_type eq "output") {
	foreach my $output (@$outputs) {
	    push @observations, &get_val($output->{$metric_name},
					 $metric_field);
	    push @trace_refs, $output->{trace_input_id};
	}
    } elsif($id_type eq "analyze") {
	foreach my $analyze (@$analyses) {
	    push @observations, &get_val($analyze->{improvement});
	    push @trace_refs, $analyze->{trace_input_id};
	}
    } else {
	die;
    }

    return (\@observations, \@trace_refs);
}

sub regress_make_predictors
{
    my ($trace_refs, $using_trace_args) = @_;
    my @matrix;

    foreach my $id (@$trace_refs) {
	push @matrix, &regress_make_predictors_trace_helper($traces_id->{$id},
							    $using_trace_args);
    }

    return \@matrix;
}

sub regress_make_predictors_trace_helper
{
    my ($trace, $using_trace_args_orig) = @_;
    my @row;
    my @using_trace_args = @$using_trace_args_orig;

    foreach my $trace_arg (@trace_args_list) {
	my $test_trace;
	my @vals; # mostly one val, but for uniform, two
	my $using_trace = shift @using_trace_args;

	$test_trace = $$traces[0];
	if(defined($test_trace->{"${trace_arg}_kind"})) {
	    @vals = &get_dist_vals($trace, $trace_arg);
	} else {
	    @vals = $trace->{$trace_arg};
	}
	die if(!defined($vals[0]));

	push @row, $vals[0] if($$using_trace[0] == true);
	if(defined($$using_trace[1])) {
	    push @row, $vals[1] if($$using_trace[1] == true);
	}

    }

    die "no unchanging row vals!" if @row == 0;

    return \@row;
}

sub regress_show_y_str
{
    my $observations = shift;
    my $first;
    my $nums_str;

    print "y = [";
    $first = true;
    foreach my $obs (@$observations) {
	my $nice_num;

	if($first) {
	    $first = false;
	} else {
	    $nums_str .= ",";
	}

	if($options_regress_log) {
	    $obs += 0.0001 if($obs == 0); # ugh. XXX
	    die "trying to take the log of an obs that's <= 0" if($obs <= 0);
	    $nice_num = sprintf "$regress_log_prec", log $obs;
	} else {
	    $nice_num = sprintf "$regress_regular_prec", $obs;
	}

	$nums_str .= $nice_num;
    }
    print $nums_str;
    print "]';\n";

    return scalar @$observations;
}

sub regress_show_X_str
{
    my $predictors = shift;
    my ($num_cols, $test_row);
    my $first;

    $test_row = $$predictors[0];
    $num_cols = @$test_row;

    print "X = [";

    $first = true;
    foreach my $row (@$predictors) {
	my $row_str;

	if($first) {
	    $first = false;
	} else {
	    print "\n";
	}

	$row_str = &regress_make_X_row_str($row);
	print $$row_str;
    }

    print "];\n";

    return (scalar @$predictors, $num_cols);
}

sub regress_make_X_row_str
{
    my $row = shift;
    my $row_str;

    $row_str = "1";
    foreach my $pred (@$row) {
	my $maybe_log_pred;
	my $power_pred;

	if($options_regress_log) {
	    $pred += 0.0001 if($pred == 0); # ugh. XXX
	    die "trying to take the log of a pred that's <= 0" if($pred <= 0);
	    $maybe_log_pred = log $pred;
	} else {
	    $maybe_log_pred = $pred;
	}
	$power_pred = 1;
	for(my $i = 0; $i < $options_regress_degree; $i++) {
	    my $nice_num;

	    $power_pred *= $maybe_log_pred;

	    if($options_regress_log) {
		$nice_num = sprintf "$regress_log_prec", $power_pred;
	    } else {
		$nice_num = sprintf "$regress_regular_prec", $power_pred;
	    }

	    $row_str .= "," . $nice_num;
	}
    }

    return \$row_str;
}

sub regress_show_y_fake_str
{
    my @fake_observations = ( 1, 3, 2 );
#    my @fake_observations = ( 1, 2, 3 );
    my $y_str_fake;

    $y_str_fake = &regress_make_vector_str(\@fake_observations, true);

    print "y = ", $y_str_fake, ";\n";

    return scalar @fake_observations;
}

# used only by fake now
sub regress_make_vector_str
{
    my $vec = shift;
    my $make_col = shift;
    my $should_make_col;

    $should_make_col = false;
    if(defined($make_col)) {
	if($make_col) {
	    $should_make_col = true;
	}
    }

    return "[" . join(",", @$vec) . "]" . ($should_make_col ? "'" : "");
}

sub regress_show_X_fake_str
{
    my @fake_predictor1 = ( 1, 1 );
    my @fake_predictor2 = ( 1, 3 );
    my @fake_predictor3 = ( 1, 2 );
    my @fake_predictors;
    my $X_str_fake;

    push @fake_predictors, \@fake_predictor1;
    push @fake_predictors, \@fake_predictor2;
    push @fake_predictors, \@fake_predictor3;

    $X_str_fake = &regress_make_matrix_str(\@fake_predictors);

    print "X = ", $X_str_fake, ";\n";

    return (scalar @fake_predictors, scalar @fake_predictor1);
}

# used only by fake now
sub regress_make_matrix_str
{
    my $matrix = shift;
    my @row_strs;

    foreach my $row (@$matrix) {
	push @row_strs, join(",", @$row);
    }

    return "[" . join("\n", @row_strs) . "]";
}

sub regress_show_alpha_str
{
    my $alpha = shift;

    print "ALPHA = $alpha;\n";
}

##########

# util for matlab

sub regress_show_num_check
{
    &regress_show_num_check_helper("B");
    &regress_show_num_check_helper("BINT");
    &regress_show_num_check_helper("R");
    &regress_show_num_check_helper("RINT");
    &regress_show_num_check_helper("STATS");
}

sub regress_show_num_check_helper
{
    my $var = shift;

    print "test_var = find(finite($var) == false);
if(length(test_var) ~= 0)
  disp('the following is the index in $var with a problem:')
  test_var
  error('NaN or Inf or -Inf in $var')
end
"
}

sub regress_show_graphs
{
    my ($num_observations, $using_trace_args) = @_;
    
    &regress_show_graph_page(1, $num_observations, $using_trace_args);
    &regress_show_graph_page(2, $num_observations, $using_trace_args);
    &regress_show_graph_page(3, $num_observations, $using_trace_args);
}

sub regress_show_graph_page
{
    my ($page, $num_observations, $using_trace_args) = @_;
    my $gfile;

    &show_new_figure;

    if($page == 1) {
	&regress_show_graph_page_1($num_observations, $using_trace_args);
    } elsif($page == 2) {
	&regress_show_graph_page_2($num_observations, $using_trace_args);
    } elsif($page == 3) {
	&regress_show_graph_page_3($num_observations, $using_trace_args);
    } else {
	die;
    }

    return if(!$options_write);
    $gfile = $options_write . ".$page.eps";
    print "print -deps $gfile;\n";
}

sub regress_show_graph_page_1
{
    my ($num_observations, $using_trace_args) = @_;

    &show_subplot(4,1,1, $regress_subplot_p1_height_correct);
    &regress_show_obs_graph($num_observations);

    &show_subplot(4,1,2, $regress_subplot_p1_height_correct);
    &regress_show_rel_error_graph($num_observations);

    &show_subplot(4,1,3, $regress_subplot_p1_height_correct);
    &regress_show_residual_graph;

    &show_subplot(4,1,4, $regress_subplot_p1_height_correct);
    &regress_show_coefficient_graph($using_trace_args);
}

sub regress_show_graph_page_2
{
    my ($num_observations, $using_trace_args) = @_;
    my $num_subplot_rows;
    my ($i, $j);
    my $used_names;
    my ($y_label_obs, $y_label_res);
    my $metric_str = &make_metric_str($options_regress_use_analyze);

    ($used_names, undef) =
	&regress_make_used_and_skipped_names($using_trace_args);
    $num_subplot_rows = @$used_names;

    $y_label_obs = $metric_str;
    $y_label_res = "residuals";

    $i = 1;
    $j = 0;
    foreach my $used_name (@$used_names) {
	my $X_col;

	$used_name =~ s/_/\\_/g;

	# 1 for the ones col.
	$X_col = 2 + $j * $options_regress_degree;
	$j++;

	&show_subplot($num_subplot_rows, 2, $i++,
		      $regress_subplot_p2_height_correct);
	&regress_show_scatter_graph($X_col, "y", $used_name, $y_label_obs,
				    true);

	&show_subplot($num_subplot_rows, 2, $i++,
		      $regress_subplot_p2_height_correct);
	&regress_show_scatter_graph($X_col, "R", $used_name, $y_label_res,
				    false);
    }
}

sub regress_show_graph_page_3
{
    my ($num_observations, $using_trace_args) = @_;

    &regress_show_norm_graph;
}

sub regress_show_scatter_graph
{
    my ($X_col, $dep, $x_label, $y_label, $trans) = @_;
    
    # gerr, can't give MarkerSize on a scatter.
    if($trans && $options_regress_log) {
	print "scatter(exp(X(:,$X_col)), exp($dep), '.');\n";
    } else {
	print "scatter(X(:,$X_col), $dep, '.');\n";
    }
    print "xlabel('$x_label');\n";
    print "ylabel('$y_label');\n";
    print "grid on;\n";
}

sub regress_show_obs_graph
{
    my $num_observations = shift;
    my $metric_str = &make_metric_str($options_regress_use_analyze);

    if($options_regress_log) {
	print "plot(exp(y), $regress_dot_marker_str);\n";
    } else {
	print "plot(y, $regress_dot_marker_str);\n";
    }
    print "set(gca,'XLim',[1,$num_observations]);\n";
    print "xlabel('observations');\n";
    print "ylabel('$metric_str');\n";
}

sub regress_show_rel_error_graph
{
    my $num_observations = shift;

    if($options_regress_log) {
	print "rel_error = abs(exp(y) - exp(X * B)) ./ exp(y);\n";
    } else {
	print "rel_error = abs(y - X * B) ./ y;\n";
    }
    print "plot(rel_error, $regress_dot_marker_str)\n";
    print "set(gca,'XLim',[1,$num_observations]);\n";
    print "xlabel('cases');\n";
    print "ylabel('abs relative error');\n";
}

sub regress_show_residual_graph
{
    print "rcoplot(R, RINT);\n";
    print "title('');\n";
    print "xlabel('cases');\n";
    print "ylabel('residuals');\n";
}

sub regress_show_coefficient_graph
{
    my $using_trace_args = shift;
    my ($used_names, $poly, $poly_len);

    print "errorbar(1:length(B), B, B - BINT(:,1), BINT(:,2) - B, 'o');\n";
    print "xlabel('coefficients');\n";
    print "ylabel('conf. of coeffs');\n";
    ($used_names, undef) =
	&regress_make_used_and_skipped_names($using_trace_args);
    ($poly, undef) = &regress_make_poly_list_of_used_names($used_names, false);
    $poly_len = @$poly;
    print "set(gca,'XTick',[1:$poly_len]);\n";
    print "set(gca,'XLim',[1,$poly_len]);\n";
    print "grid on;\n";
}

sub regress_show_norm_graph
{
    my $metric_str = &make_metric_str($options_regress_use_analyze);

    print "normplot(y);\n";
    print "xlabel('$metric_str');\n";
    print "ylabel('probability');\n";
    print "title('Normal probability plot');\n";
}

######################################################################

sub conf_ints
{
    my %collected_data; # hash of list refs, id'd by trace_input_id
    my %collected_stats; # hash of stat refs, id'd by trace_input_id
    my $sorted_traces;

    &set_behavior;
    &conf_ints_checks;

    foreach my $output (@$outputs) {
	my $val = &get_val($output->{$metric_name}, $metric_field);
	my $list;

	$list = $collected_data{$output->{trace_input_id}};
	push @$list, $val;
	$collected_data{$output->{trace_input_id}} = $list;
    }

    foreach my $key (keys %collected_data) {
	my $list = $collected_data{$key};
	my $stats;

	$stats = &stat_nums_compute($list);
	$collected_stats{$key} = $stats;
    }

    $sorted_traces = [sort conf_ints_sort_callback keys %collected_stats];

    &conf_ints_show_header;

    if($options_graph) {
	&conf_ints_show_graph($sorted_traces, \%collected_stats);
    } else {
	&conf_ints_show_non_graph($sorted_traces, \%collected_stats);
    }
}

sub conf_ints_show_graph
{
    my ($sorted_traces, $collected_stats) = @_;
    my $kind;

    &show_new_figure;
    $kind = &generic_show_graph_x($sorted_traces);

    if($options_graph_type eq "bar") {
	&conf_ints_show_graph_y($sorted_traces, $collected_stats);
	&conf_ints_show_graph_err($sorted_traces, $collected_stats);
	print "bar(x,y,'y');\n";
	print "hold;\n";
	# gerr, can't give MarkerSize on an errorbar.
	print "errorbar(x,y,err,'ok');\n";
	print "hold;\n";
    } elsif($options_graph_type eq "line") {
	die;
    } else {
	die;
    }

    # skipping the legend
    &conf_ints_show_graph_labels($kind);
    &show_box_off;
    &generic_show_graph_write;
}

sub conf_ints_show_graph_y
{
    my ($sorted_traces, $collected_stats) = @_;
    my $first;

    print "y = [";
    $first = true;
    foreach my $trace_id (@$sorted_traces) {
	my $sample_mean;

	$sample_mean = $collected_stats->{$trace_id}{sample_mean};

	if($first) {
	    $first = false;
	} else {
	    print ",";
	}
	
	print "$sample_mean";
    }
    print "];\n";
}

sub conf_ints_show_graph_err
{
    my ($sorted_traces, $collected_stats) = @_;
    my $first;

    print "err = [";
    $first = true;
    foreach my $trace_id (@$sorted_traces) {
	my $conf_delta_95;

	$conf_delta_95 = $collected_stats->{$trace_id}{conf_delta_95};

	if($first) {
	    $first = false;
	} else {
	    print ",";
	}
	
	print "$conf_delta_95";
    }
    print "];\n";
}

sub conf_ints_show_graph_labels
{
    my $kind = shift;

    &generic_show_graph_labels($kind, false);
}

sub conf_ints_show_non_graph
{
    my ($sorted_traces, $collected_stats) = @_;

    foreach my $trace_id (@$sorted_traces) {
	my ($sample_mean, $conf_delta_95);

	$sample_mean = $collected_stats->{$trace_id}{sample_mean};
	$conf_delta_95 = $collected_stats->{$trace_id}{conf_delta_95};

	print "trace_id: $trace_id, mean: $sample_mean, " .
	  "conf_delta_95: $conf_delta_95\n";
    }
}

sub conf_ints_show_header
{
    my @header_lines;

    @header_lines = &conf_ints_make_header;
    # send lines to stdout
    foreach my $line (@header_lines) {
	print "$options_comment_char $line";
    }
    print "\n";

    if($options_graph) {
	# add some stuff b4hand:
	print "disp('behavior: $options_behavior')\n";
	# send munged lines to stdout that will tell matlab to display lines.
	foreach my $line (@header_lines) {
	    chop $line;
	    print "disp('$line')\n";
	}
	print "\n";
    }
}

sub conf_ints_make_header
{
    my @lines;

    push @lines, &show_header_trace_range_helper;

    return @lines;
}

sub conf_ints_sort_callback
{
    my ($set1_trace_input_id, $set2_trace_input_id) = ($a, $b);

    return &sort_on_changing_behavior_helper($set1_trace_input_id,
					     $set2_trace_input_id);
}

sub conf_ints_checks
{
    # check that all the outputs use the same scheduler
    &check_same_sched($outputs);
    
    # ensure all dists have the same kind across traces and ensure
    # that only the specified trace behavior changes
    &check_dist_kind_and_behavior_changes;
}

######################################################################

sub improve_stats
{
    my @improvements;
    my $count;
    my @cumul;
    my ($baseline_sched_type, $new_sched_type);

    # should rename cumul_checks to something generic
    ($baseline_sched_type, $new_sched_type) = &cumul_checks;

    # actual computation start
    foreach my $analyze (@$analyses) {
	my $improvement;

	$improvement = $analyze->{improvement};
	push @improvements, $improvement;
    }

#    $stats = &stat_nums_compute(\@improvements);
    &stat_nums_print(\@improvements);
}

######################################################################

sub sru_hack
{
    foreach my $output (@$outputs) {
	my $val = &get_val($output->{scaled_res_usage}, "mean");
	print "$val\n";
    }
}

######################################################################

sub deadlines
{
    foreach my $output (@$outputs) {
	my $val = &get_val($output->{deadlines_met}, "total");
	print "$val\n";
    }
}

######################################################################

sub varvrt
{
    foreach my $output (@$outputs) {
	my $val = &get_val($output->{vrt}, "stddev");
	print "$val\n";
    }
}

######################################################################

sub varurr
{
    foreach my $output (@$outputs) {
	my $val;

	if($options_take_req) {
	    $val = &get_val($output->{req_res_usage}, "stddev");
	} else {
	    $val = &get_val($output->{all_res_usage}, "stddev");
	}
	print "$val\n";
    }
}

######################################################################

sub sbrcomp
{
    my %collected_data; # hash of list refs, id'd by trace_input_id
    my %collected_stats; # hash of stat refs, id'd by trace_input_id
    my $sorted_traces;

    &set_behavior;
    &conf_ints_checks;

    die if($options_metric ne "scaled_res_usage:mean");

    foreach my $output (@$outputs) {
	my $val = &get_val($output->{$metric_name}, $metric_field);
	my $list;

	$list = $collected_data{$output->{trace_input_id}};
	die if(defined $list);
	push @$list, $val;
	$collected_data{$output->{trace_input_id}} = $list;
    }

    foreach my $key (keys %collected_data) {
	my $list = $collected_data{$key};

	$collected_stats{$key} = $$list[0];
    }

    $sorted_traces = [sort conf_ints_sort_callback keys %collected_stats];

    &conf_ints_show_header;

    if($options_graph) {
	&sbrcomp_show_graph($sorted_traces, \%collected_stats);
    } else {
	die;
    }
}

sub sbrcomp_show_graph
{
    my ($sorted_traces, $collected_stats) = @_;
    my $kind;

    &show_new_figure;
    $kind = &generic_show_graph_x($sorted_traces);

    &sbrcomp_show_graph_y($sorted_traces, $collected_stats);
    if($options_graph_type eq "bar") {
	print "bar(x,y);\n";
    } elsif($options_graph_type eq "line") {
	print "plot(x,y,'-xk');\n";
    } else {
	die;
    }

    # skipping the legend
#    &conf_ints_show_graph_labels($kind);
    &generic_show_graph_labels($kind, false);
    print "grid on;\n";
    &show_box_off;
    &generic_show_graph_write;
}

sub sbrcomp_show_graph_y
{
    my ($sorted_traces, $collected_stats) = @_;
    my $first;

    print "y = [";
    $first = true;
    foreach my $trace_id (@$sorted_traces) {
	my $val;

	$val = $collected_stats->{$trace_id};

	if($first) {
	    $first = false;
	} else {
	    print ",";
	}
	
	print "$val";
    }
    print "];\n";
}

######################################################################

sub both_dep
{
    my ($x_metric, $y_metric);
    my ($x_metric_dir, $y_metric_dir);
    my ($x_metric_full_name, $y_metric_full_name);
    my ($x_metric_name, $y_metric_name);
    my ($x_metric_field, $y_metric_field);
    my @points;

    &set_behavior;
    &conf_ints_checks;

    if($options_both_dep_deps =~ /([^,]+),([^,]+)/) {
	$x_metric = $1;
	$y_metric = $2;
    } else {
	die;
    }

    if($x_metric =~ /^([+-])(.+)$/) {
	$x_metric_dir = $1;
	$x_metric_full_name = $2;
    } else {
	die;
    }

    if($y_metric =~ /^([+-])(.+)$/) {
	$y_metric_dir = $1;
	$y_metric_full_name = $2;
    } else {
	die;
    }

    if($x_metric_full_name =~ /^([^:]+):([^:]+)$/) {
	($x_metric_name, $x_metric_field) = ($1, $2);
    } else {
	$x_metric_name = $x_metric_full_name;
    }

    if($y_metric_full_name =~ /^([^:]+):([^:]+)$/) {
	($y_metric_name, $y_metric_field) = ($1, $2);
    } else {
	$y_metric_name = $y_metric_full_name;
    }

    foreach my $output (@$outputs) {
	my ($x_val, $y_val);

	$x_val = &get_val($output->{$x_metric_name}, $x_metric_field);
	$y_val = &get_val($output->{$y_metric_name}, $y_metric_field);

	push @points, [$x_val, $y_val];

    }

    $both_dep_closure_hack_x_dir = $x_metric_dir;
    @points = sort both_dep_sort_callback @points;

    foreach my $point (@points) {
	my ($x_val, $y_val);

	$x_val = $$point[0];
	$y_val = $$point[1];

	print "$x_val, $y_val\n";
    }
}

sub both_dep_sort_callback
{
    my ($point1, $point2) = ($a, $b);
    my ($comp_point1, $comp_point2);

    $comp_point1 = $$point1[0];
    $comp_point2 = $$point2[0];

    if($comp_point1 == $comp_point2) {
	# reasonable...  hopefully the direction is correct for this! XXX
	$comp_point1 = $$point1[1];
	$comp_point2 = $$point2[1];
    }

    if($both_dep_closure_hack_x_dir eq "+") {
	return $comp_point1 <=> $comp_point2;
#	return $$point1[0] <=> $$point2[0];
    } elsif($both_dep_closure_hack_x_dir eq "-") {
	return $comp_point2 <=> $comp_point1;
#	return $$point2[0] <=> $$point1[0];
    } else {
	die;
    }
}

######################################################################

sub ttst
{
    my @points;

    foreach my $output (@$outputs) {
	my ($tt_val, $st_val, $metric_val);
	my $tt_over_st;
	my $trace;

	# find trace id for this output.

        # say it's $trace.

	$trace = $traces_id->{$output->{"trace_input_id"}};

	die if($trace->{"service_time_kind"} ne "exponential");
	die if($trace->{"think_time_kind"} ne "exponential");

	$st_val = $trace->{"service_time_mean"};
	$tt_val = $trace->{"think_time_mean"};
	$metric_val = &get_val($output->{$metric_name}, $metric_field);

	$tt_over_st = $tt_val / $st_val;

	push @points, [$tt_over_st, $metric_val];
    }

    @points = sort ttst_sort_callback @points;

    foreach my $point (@points) {
	my ($x_val, $y_val);

	$x_val = $$point[0];
	$y_val = $$point[1];

	print "$x_val, $y_val\n";
    }
}

sub ttst_sort_callback
{
    my ($point1, $point2) = ($a, $b);

    return $$point1[0] <=> $$point2[0];
}

######################################################################

sub output_metric
{
    my $y_metric;
    my $y_metric_dir;
    my $y_metric_full_name;
    my $y_metric_name;
    my $y_metric_field;
    my @points;

    &set_behavior;
    &conf_ints_checks;

    $y_metric_name = $metric_name;
    $y_metric_field = $metric_field;

    foreach my $output (@$outputs) {
	my ($x_val, $y_val);

	$x_val = $output->{trace_input_id};
	$y_val = &get_val($output->{$y_metric_name}, $y_metric_field);

	push @points, [$x_val, $y_val];

    }

    @points = sort output_metric_sort_callback @points;

    foreach my $point (@points) {
	my ($x_val, $y_val);
	my $trace;
	my $dist_val;
	my $val;
	my $test_kind;

	$x_val = $$point[0];
	$trace = $traces_id->{$x_val};

	$test_kind = $trace->{"${from_opt_behavior}_kind"};
	if(defined($test_kind)) {
	    my ($val_a, $val_b) = &get_dist_vals($trace, $from_opt_behavior);

	    if(defined($val_b)) {

		$val = ($val_a + $val_b) / 2;
	    } else {
		
		$val = $val_a;
	    }
	} else {
	    $val = $trace->{$from_opt_behavior};
	}
	
	$y_val = $$point[1];

	print "$val, $y_val\n";
    }
}

sub output_metric_sort_callback
{
    my ($point1, $point2) = ($a, $b);
    my ($trace_input_id1, $trace_input_id2);

    $trace_input_id1 = $$point1[0];
    $trace_input_id2 = $$point2[0];

    return &sort_on_changing_behavior_helper($trace_input_id1,
					     $trace_input_id2);
}

######################################################################

sub uncharged_load
{
    my $y_metric;
    my $y_metric_dir;
    my $y_metric_full_name;
    my $y_metric_name;
    my $y_metric_field;
    my @points;

    &set_behavior;
    &conf_ints_checks;

    $y_metric_name = $metric_name;
    $y_metric_field = $metric_field;

    foreach my $output (@$outputs) {
	my ($x_val, $req_load, $total_load, $uncharged_load);

	$x_val = $output->{trace_input_id};
	$total_load = &get_val($output->{load}, undef);
	$req_load = &get_val($output->{req_load}, undef);

	$uncharged_load = $total_load - $req_load;

	push @points, [$x_val, $uncharged_load];

    }

    @points = sort output_metric_sort_callback @points;

    foreach my $point (@points) {
	my ($x_val, $y_val);
	my $trace;
	my $dist_val;
	my $val;
	my $test_kind;

	$x_val = $$point[0];
	$trace = $traces_id->{$x_val};

	$test_kind = $trace->{"${from_opt_behavior}_kind"};
	if(defined($test_kind)) {
	    my ($val_a, $val_b) = &get_dist_vals($trace, $from_opt_behavior);

	    if(defined($val_b)) {

		$val = ($val_a + $val_b) / 2;
	    } else {
		
		$val = $val_a;
	    }
	} else {
	    $val = $trace->{$from_opt_behavior};
	}
	
	$y_val = $$point[1];

	print "$val, $y_val\n";
    }
}

######################################################################

sub check_same_sched
{
    my $local_outputs = shift;
    my $first_sched_str;
    my $first;

    $first = true;
    foreach my $output (@$local_outputs) {
	my $sched_str;

	$sched_str = &make_sched_str($output);

	if($first) {
	    $first = false;
	    $first_sched_str = $sched_str;
	    next;
	}

	die "sched types are supposed to be the same"
	    if($sched_str ne $first_sched_str);
    }

    return $first_sched_str;
}

sub set_behavior
{
    die "must set behavior" if(!$options_behavior);
    if($options_behavior =~ /^([+-])(.+)$/) {
	$from_opt_behavior_dir = $1;
	$from_opt_behavior = $2;

	die "unknown behavior"
	    unless defined($trace_args{$from_opt_behavior});
    } else {
	die "couldn't parse ba-compare-behavior";
    }
}

sub check_dist_kind_and_behavior_changes
{
    foreach my $trace_arg (@trace_args_list) {
	my $test_trace;
	my $test_kind;

	$test_trace = $$traces[0];
	$test_kind = $test_trace->{"${trace_arg}_kind"};

	if(defined($test_kind)) {
	    my @test_vals = &get_dist_vals($test_trace, $trace_arg);

	    &check_ensure_same_dist($trace_arg, \@test_vals);
	    &check_good_dist_kind($trace_arg);
	} else {
	    my $test_val = $test_trace->{$trace_arg};

	    &check_ensure_same_val($trace_arg, $test_val);
	}
    }
}

sub check_ensure_same_dist
{
    my ($trace_arg, $test_vals) = @_;

    return if $trace_arg eq $from_opt_behavior;

    foreach my $trace (@$traces) {
	my ($val1, $val2) = &get_dist_vals($trace, $trace_arg);

	die "dist val changes when it shouldn't" if($val1 != $$test_vals[0]);
	if(defined($$test_vals[1])) {
	    die "dist val changes when it shouldn't"
		if($val2 != $$test_vals[1]);
	}
    }
}

sub check_ensure_same_val
{
    my ($trace_arg, $test_val) = @_;

    return if $trace_arg eq $from_opt_behavior;

    foreach my $trace (@$traces) {
	my $val = $trace->{$trace_arg};

	if($test_val != $val) {
	    print STDERR "test_val: $test_val, val: $val\n";
	    die "val changes when it shouldn't";
	}
    }
}

sub generic_show_graph_x
{
    my $list_of_trace_ids = shift;
    my $first;
    my $kind;

    print "x = [";
    $first = true;
    foreach my $trace_input_id (@$list_of_trace_ids) {
	my $trace;
	my (@vals, $val);

	$trace = $traces_id->{$trace_input_id};
	$kind = $trace->{"${from_opt_behavior}_kind"};

	if($first) {
	    $first = false;
	} else {
	    print ",";
	}

	if(defined($trace->{"${from_opt_behavior}_kind"})) {
	    @vals = &get_dist_vals($trace, $from_opt_behavior);
	} else {
	    @vals = $trace->{$from_opt_behavior};
	}
	die if(!defined($vals[0]));

	if($vals[1]) {
	    $val = ($vals[0] + $vals[1]) / 2;
	} else {
	    $val = $vals[0];
	}

	print "$val";
    }
    print "];\n";

    return $kind;
}

sub sort_on_changing_behavior_helper
{
    my ($set1_trace_input_id, $set2_trace_input_id) = @_;
    my ($trace1, $trace2);
    my $test_kind;
    my ($val1, $val2);
    $trace1 = $traces_id->{$set1_trace_input_id};
    $trace2 = $traces_id->{$set2_trace_input_id};

    $test_kind = $trace1->{"${from_opt_behavior}_kind"};
    if(defined($test_kind)) {
	my ($val_1a, $val_1b) = &get_dist_vals($trace1, $from_opt_behavior);
	my ($val_2a, $val_2b) = &get_dist_vals($trace2, $from_opt_behavior);

	if(defined($val_1b)) {
	    die "can't be the same vals" if $val_1a == $val_2a &&
		$val_1b == $val_2b;

	    $val1 = ($val_1a + $val_1b) / 2;
	    $val2 = ($val_2a + $val_2b) / 2;
	} else {
	    die "can't be the same val" if $val_1a == $val_2a;

	    $val1 = $val_1a;
	    $val2 = $val_2a;
	}
    } else {
	$val1 = $trace1->{$from_opt_behavior};
	$val2 = $trace2->{$from_opt_behavior};
	
	die "can't be the same val" if $val1 == $val2;
    }

    if($from_opt_behavior_dir eq "+") {
	return $val1 <=> $val2;
    } else {
	return $val2 <=> $val1;
    }
}

sub sort_numerically_callback
{
    $a <=> $b;
}

sub check_good_dist_kind
{
    my $trace_arg = shift;
    my $first;
    my $first_dist_kind;

    $first = true;
    foreach my $trace (@$traces) {
	my $dist_kind = $trace->{"${trace_arg}_kind"};

	if($first) {
	    $first = false;
	    $first_dist_kind = $dist_kind;
	    next;
	}

	die "dist kinds are supposed to be the same"
	    if($dist_kind ne $first_dist_kind);
    }
}

sub check_one_type_of_metric_for_analyze
{
    my ($first_metric_order, $first_metric_name, $first_metric_field);
    my $first;

    $first = true;
    foreach my $analyze (@$analyses) {
	if($first) {
	    $first = false;
	    $first_metric_order = $analyze->{metric_order};
	    $first_metric_name = $analyze->{metric_name};
	    $first_metric_field = $analyze->{metric_field};
	    next;
	}

	die "metrics not the same"
	    if($first_metric_order ne $analyze->{metric_order} ||
	       $first_metric_name ne $analyze->{metric_name});
	if(defined($first_metric_field)) {
	    die "metrics not the same"
		if(!defined($analyze->{metric_field}));
	    die "metrics not the same"
		if($first_metric_field ne $analyze->{metric_field});
	}
    }

    return ($first_metric_order, $first_metric_name, $first_metric_field);
}

sub get_sched_type
{
    my $output = shift;

    if($output->{dis_sched} eq "NOOP") {
	return "interactive";
    } elsif($output->{dis_sched} eq "n/a") {
	return "batch";
    } else {
	return "batchactive";
    }
}

sub make_list_into_hash_by_id
{
    my ($list, $hash) = @_;

    foreach my $elem (@$list) {
	my $id = $elem->{id};

	$hash->{$id} = $elem;
    }
}

sub show_commented_trace_ranges
{
    my $trace_ranges_lines;

    print "$options_comment_char trace range info:\n";
    $trace_ranges_lines = &trace_ranges_helper;
    foreach my $line (@$trace_ranges_lines) {
	print "$options_comment_char " . $line;
    }
}

sub get_dist_vals
{
    my ($trace, $trace_arg) = @_;
    my ($kind, $alpha, $beta, $mean, $prob);

    $kind = $trace->{"${trace_arg}_kind"};
    $alpha = $trace->{"${trace_arg}_alpha"};
    $beta = $trace->{"${trace_arg}_beta"};
    $mean = $trace->{"${trace_arg}_mean"};
    $prob = $trace->{"${trace_arg}_prob"};

    if($kind eq "exponential") {
	return $mean;
    } elsif($kind eq "pareto") {
	return $mean; # this is the pareto's alpha
    } elsif($kind eq "uniform") {
	my $lower = $alpha;
	my $upper = $beta;

	if($options_regress_trans_tscr) {
	    my $avg;
	    my $distance;
	
	    $avg = ($lower + $upper) / 2;
	    $distance = $upper - $lower;
	    return ($avg, $distance);
	} else {
	    return ($lower, $upper);
	}
    } elsif($kind eq "bernoulli") {
	return $prob;
    } elsif($kind eq "poisson") {
	return $mean;
    } else {
	die;
    }
}

sub show_header_trace_range_helper
{
    my @lines;
    my $trace_ranges_lines;

    push @lines, "trace range info:\n";
    $trace_ranges_lines = &trace_ranges_helper;
    foreach my $trace_range_line (@$trace_ranges_lines) {
	$trace_range_line = "    " . $trace_range_line;
    }
    push @lines, @$trace_ranges_lines;

    return @lines;
}

sub show_new_figure
{
    my ($axis_length_end, $axis_height_end);

    $axis_length_end = $image_length - $axis_length_offset;
    $axis_height_end = $image_height - $axis_height_offset;

    print "figure;\n";
    print "set(gca,'FontSize',$font_size);\n";
    print "set(gca,'Units','inches');\n";
    # [left bottom width height]
    print "set(gca,'Position',[$axis_length_offset $axis_height_offset " .
	"$axis_length_end $axis_height_end]);\n";
}

sub show_box_off
{
    print "box off;\n";
}

sub show_subplot
{
    my ($rows, $cols, $index, $height_correct) = @_;

    print "subplot($rows, $cols, $index);\n";
    print "set(gca,'FontSize',$font_size);\n";
    if($height_correct) {
	print "hcor = get(gca, 'Position');\n";
	print "set(gca, 'Position', [hcor(1), hcor(2), hcor(3), " .
	    "hcor(4) * $height_correct]);\n";
    }
}

sub make_metric_str
{
    my $improve = shift;
    my $metric_str;

    if($metric_field) {
	$metric_str = "$metric_name ($metric_field)";
    } else {
	$metric_str = "$metric_name";
    }

    if($improve) {
	$metric_str .= " improvement";
    }
    
    return $metric_str;
}

sub generic_show_graph_labels
{
    my ($kind, $should_show_improve) = @_;
    my $metric_str = &make_metric_str($should_show_improve);
    my $nice_metric_str = &make_nice_metric_str($metric_str);
    my $nice_behavior_str = &make_nice_behavior_str($from_opt_behavior, $kind);

    print "xlabel('$nice_behavior_str','FontSize',$font_size);\n";
    print "ylabel('$nice_metric_str','FontSize',$font_size);\n";
}

sub make_nice_behavior_str
{
    my ($behavior, $kind) = @_;

    if($kind) {
	if($from_opt_behavior eq "service_time" && $kind eq "exponential") {
	    return "service time (\\mu of exp. dist.)";
	}
	if($from_opt_behavior eq "think_time" && $kind eq "exponential") {
	    return "think time (\\mu of exp. dist.)";
	}
	if($from_opt_behavior eq "service_time" && $kind eq "pareto") {
	    return "service time (\\alpha of pareto dist.)";
	}
	if($from_opt_behavior eq "think_time" && $kind eq "pareto") {
	    return "think time (\\alpha of pareto dist.)";
	}
	if($from_opt_behavior eq "task_set_change_range" &&
	   $kind eq "uniform") {
	    return "task set change prob. (midpoint of uni. dist.)";
	}
	if($from_opt_behavior eq "tasks_per_task_set" && $kind eq "uniform") {
	    return "tasks per task set (midpoint of uni. dist.)";
	}
    }

    if($behavior eq "num_users") {
	return "number of users";
    }

    $behavior =~ s/_/\\_/g;
    if($kind) {
	return "$behavior ($kind)";
    } else {
	return $behavior;
    }
}

sub make_nice_metric_str
{
    my $metric_str = shift;

    if($metric_str eq "vrt (mean)") {
	return "visible response time (\\mu)";
    }
    if($metric_str eq "vsd (mean)") {
	return "visible slowdown (\\mu)";
    }
    if($metric_str eq "vrt (count)") {
	return "number of tasks";
    }

    $metric_str =~ s/_/\\_/g;

    return $metric_str;
}

sub generic_show_graph_write
{
    my $gfile;

#    print "set(gcf,'PaperPosition',[0 0 $image_length $image_height]);\n";
    return unless $options_write;
    $gfile = $options_write . ".eps";
    print "print -deps $gfile;\n";
}

######################################################################

sub parse_options
{
    my ( %opt_hash );

    $Getopt::Long::autoabbrev = false;
    $Getopt::Long::order = $REQUIRE_ORDER;
    $Getopt::Long::ignorecase = false;

    $options_db_host = $db_default_loc;

    if(GetOptions(\%opt_hash,
		  "show-settings",
		  "db-host=s",
		  "comment-char=s",
		  "metric=s",
		  "graph",
		  "graph-type=s",
		  "write=s",
		  "behavior=s",

		  # dump-analyze opts:

		  # cumul opts:

		  # trace-range opts:

		  # compare-ba opts:
		  "compare-ba-improve",

		  # regress opts:
		  "regress-use-analyze",
		  "regress-ci=s",
		  "regress-degree=s",
		  "regress-log",
		  "regress-trans-tscr",

		  # conf-ints opts:
		  # improve-stats opts:
		  # sru-hack opts:
		  # deadlines opts:
		  # varvrt opts:
		  # varurr opts:
		  "take-req",
		  # sbrcomp opts:
		  # ttst opts:
		  # output-metric opts:
		  # uncharged-load opts:
		  # both-dep opts:
                  "both-dep-deps=s"
		  ) == false) {
	die "bad option(s): error in GetOptions()";
    }

    $options_command = $ARGV[0];
    if(defined($opt_hash{"db-host"})) {
	$options_db_host = $opt_hash{"db-host"};
    }
    if(defined($opt_hash{"comment-char"})) {
	$options_comment_char = $opt_hash{"comment-char"};
    }
    if(defined($opt_hash{"metric"})) {
	$options_metric = $opt_hash{"metric"};
    }
    if(defined($opt_hash{"graph"})) {
	$options_graph = true;
    }
    if(defined($opt_hash{"graph-type"})) {
	$options_graph_type = $opt_hash{"graph-type"};
    }
    if(defined($opt_hash{"write"})) {
	$options_write = $opt_hash{"write"};
    }
    # dump-analyze opts:
    # cumul opts:
    # trace-range opts:
    # compare-ba opts:
    if(defined($opt_hash{"compare-ba-improve"})) {
	$options_compare_ba_improve = true;
    }
    if(defined($opt_hash{"behavior"})) {
	$options_behavior = $opt_hash{"behavior"};
    }
    # regress opts:
    if(defined($opt_hash{"regress-use-analyze"})) {
	$options_regress_use_analyze = true;
    }
    if(defined($opt_hash{"regress-ci"})) {
	$options_regress_ci = $opt_hash{"regress-ci"};
    }
    if(defined($opt_hash{"regress-degree"})) {
	$options_regress_degree = $opt_hash{"regress-degree"};
    }
    if(defined($opt_hash{"regress-log"})) {
	$options_regress_log = true;
    }
    if(defined($opt_hash{"regress-trans-tscr"})) {
	$options_regress_trans_tscr = true;
    }
    # conf-ints opts:
    # improve-stats opts:
    # sru-hack opts:
    # deadlines opts:
    # varvrt opts:
    # varurr opts:
    if(defined($opt_hash{"take-req"})) {
	$options_take_req = true;
    }
    # sbrcomp opts:
    # ttst opts:
    # output-metric opts:
    # uncharged-load opts:
    # both-dep opts:
    if(defined($opt_hash{"both-dep-deps"})) {
	$options_both_dep_deps = $opt_hash{"both-dep-deps"};
    }

    &show_settings if(defined($opt_hash{"show-settings"}));

    die "bad option(s): no command" if(@ARGV != 1);
    die "bad option(s): left over args" if(@ARGV > 1);

    if($options_command eq "dump-analyze") {
	# no options to check
    } elsif($options_command eq "cumul") {
	# no options to check
    } elsif($options_command eq "trace-ranges") {
	# no options to check
    } elsif($options_command eq "compare-ba") {
	die "unknown graph type" if($options_graph_type ne "bar" &&
				    $options_graph_type ne "line");
    } elsif($options_command eq "regress") {
	die "regress-ci must be between 0 and 100"
	    if($options_regress_ci < 0 || $options_regress_ci > 100);
	die "regree-degree must be greater than 0"
	    if($options_regress_degree < 1);
    } elsif($options_command eq "conf-ints") {
	# no options to check
    } elsif($options_command eq "improve-stats") {
	# no options to check
    } elsif($options_command eq "sru-hack") {
	# no options to check
    } elsif($options_command eq "deadlines") {
	# no options to check
    } elsif($options_command eq "varvrt") {
	# no options to check
    } elsif($options_command eq "varurr") {
	# no options to check
    } elsif($options_command eq "sbrcomp") {
	# no options to check
    } elsif($options_command eq "both-dep") {
        die if(!defined($options_both_dep_deps));
        die if(!defined($options_behavior));
	# no options to check
    } elsif($options_command eq "ttst") {
	# no options to check
    } elsif($options_command eq "output-metric") {
	# no options to check
    } elsif($options_command eq "uncharged-load") {
	# no options to check
    } else {
	die "unknown command `$options_command'";
    }

    if($options_command ne "dump-analyze") {
	# no options to check that shouldn't be set
    }
    if($options_command ne "cumul") {
	# no options to check that shouldn't be set
    }
    if($options_command ne "trace-ranges") {
	# no options to check that shouldn't be set
    }
    if($options_command ne "compare-ba") {
	die "can't set compare-ba-improve when not compare-ba'ing"
	    if(defined($opt_hash{"compare-ba-improve"}));
    }
    if($options_command ne "regress") {
	die "can't set regress-use-analyze when not regress'ng"
	    if(defined($opt_hash{"regress-use-analyze"}));
	die "can't set regress-ci when not regress'ing"
	    if(defined($opt_hash{"regress-ci"}));
	die "can't set regress-degree when not regress'ing"
	    if(defined($opt_hash{"regress-degree"}));
	die "can't set regress-log when not regress'ing"
	    if(defined($opt_hash{"regress-log"}));
	die "can't set regress-trans-tscr when not regress'ing"
	    if(defined($opt_hash{"regress-trans-tscr"}));
    }
    if($options_command ne "conf-ints") {
	# no options to check that shouldn't be set
    }
    if($options_command ne "improve-stats") {
	# no options to check that shouldn't be set, at least, i don't think
    }
    if($options_command ne "sru-hack") {
	# no options to check that shouldn't be set, at least, i don't think
    }
    if($options_command ne "deadlines") {
	# no options to check that shouldn't be set, at least, i don't think
    }
    if($options_command ne "varvrt") {
	# no options to check that shouldn't be set, at least, i don't think
    }
    if($options_command ne "varurr") {
	# no options to check that shouldn't be set, at least, i don't think
    }
    if($options_command ne "sbrcomp") {
	# no options to check that shouldn't be set, at least, i don't think
    }
    if($options_command ne "both-dep") {
	# no options to check that shouldn't be set, at least, i don't think
    }
    if($options_command ne "ttst") {
	# no options to check that shouldn't be set, at least, i don't think
    }
    if($options_command ne "output-metric") {
	# no options to check that shouldn't be set, at least, i don't think
    }
    if($options_command ne "uncharged-load") {
	# no options to check that shouldn't be set, at least, i don't think
    }

    # dualies
    if($options_command ne "cumul" && $options_command ne "compare-ba" &&
       $options_command ne "conf-ints" && $options_command ne "sbrcomp"
       && $options_command ne "ttst") {
	die "can't set graph when not cumul'ing or compare-ba'ing\n"
	    if(defined($opt_hash{"graph"}));
    }
}

sub show_settings
{
    my $str;

    $str = "";
    $str .= &show_settings_helper;

    # so it doesn't interfere with the output...
    print STDERR "$str";
    print STDERR "\n";
}

sub show_settings_helper
{
    my $str;

    $str = "";

    $str .= "db host: $options_db_host\n";
    $str .= "command: $options_command\n";
    $str .= "comment-char: $options_comment_char\n";
    if($options_graph) {
	$str .= "graph: true\n";
    } else {
	$str .= "graph: false\n";
    }
    if(defined($options_metric)) {
	$str .= "metric: $options_metric\n";
    }
    if(defined($options_write)) {
	$str .= "write: $options_write\n";
    } else {
	$str .= "write: (none)\n";
    }
    if($options_command eq "dump-analyze") {
	# no options
    } elsif($options_command eq "cumul") {
	# no options
    } elsif($options_command eq "trace-ranges") {
	# no options
    } elsif($options_command eq "compare-ba") {
	if($options_compare_ba_improve) {
	    $str .= "compare-ba-improve: true\n";
	} else {
	    $str .= "compare-ba-improve: false\n";
	}
	$str .= "behavior: $options_behavior\n" if(defined($options_behavior));
	$str .= "graph-type: $options_graph_type\n";
    } elsif($options_command eq "regress") {
	if($options_regress_use_analyze) {
	    $str .= "regress-use-analyze: true\n";
	} else {
	    $str .= "regress-use-analyze: false\n";
	}
	$str .= "regress-ci: $options_regress_ci\n";
	$str .= "regress-degree: $options_regress_degree\n";
	if($options_regress_log) {
	    $str .= "regress-log: true\n";
	} else {
	    $str .= "regress-log: false\n";
	}
	if($options_regress_trans_tscr) {
	    $str .= "regress-tscr: true\n";
	} else {
	    $str .= "regress-tscr: false\n";
	}
    } elsif($options_command eq "conf-ints") {
	$str .= "behavior: $options_behavior\n" if(defined($options_behavior));
	$str .= "graph-type: $options_graph_type\n";
    } elsif($options_command eq "improve-stats") {
	# no options
    } elsif($options_command eq "sru-hack") {
	# no options
    } elsif($options_command eq "deadlines") {
	# no options
    } elsif($options_command eq "varvrt") {
	# no options
    } elsif($options_command eq "varurr") {
	if($options_take_req) {
	    $str .= "take-req: true\n";
	} else {
	    $str .= "take-req: false\n";
	}
    } elsif($options_command eq "sbrcomp") {
	# no options
    } elsif($options_command eq "both-dep") {
	$str .= "both-dep-deps: $options_both_dep_deps\n";
    } elsif($options_command eq "ttst") {
	# no options
    } elsif($options_command eq "output-metric") {
	# no options
    } elsif($options_command eq "uncharged-load") {
	# no options
    } else {
	die;
    }

    return $str;
}

######################################################################
