#include <ba_config.h>
#include <ba_sim.h>

struct ba_task_s {
    ba_id_t id;

    ba_bool valid_service_delta_time;
    ba_real service_delta_time;
    ba_bool valid_think_delta_time;
    ba_real think_delta_time;
    ba_bool valid_will_be_requested;
    ba_bool will_be_requested;

    ba_bool counted_in_warmup;
    ba_real completed_before_warmup;
    ba_real st_after_warmup;

    ba_real completed_delta_time;
    ba_real completed_warmup_delta_time;

    // the bools just refer to if the times are valid (that is, if
    // they have been set).  so, something can be executed and
    // canceled, or disclosed and requested.  the latter just means
    // that the task was disclosed (and you can find out when by
    // looking at the disclosed time) and was (later) requested.

    ba_bool running;

    ba_bool disclosed;
    ba_real disclosed_abs_time;
    ba_bool executed;
    ba_real executed_abs_time;
    ba_bool requested;
    ba_real requested_abs_time;
    ba_bool finished;
    ba_real finished_abs_time;
    ba_bool canceled;
    ba_real canceled_abs_time;

    ba_task_set_t *task_set;
};

extern ba_id_t
ba_task_id_get(ba_task_t *task)
{
    BA_ASSERT(task);

    return task->id;
}

extern ba_bool
ba_task_counted_in_warmup_get(ba_task_t *task,
			      ba_real *completed_before_warmup_out,
			      ba_real *st_after_warmup_out)
{
    BA_ASSERT(task);

    if(task->counted_in_warmup) {
	if(completed_before_warmup_out) {
	    *completed_before_warmup_out = task->completed_before_warmup;
	}
	if(st_after_warmup_out) {
	    *st_after_warmup_out = task->st_after_warmup;
	}
    }

    return task->counted_in_warmup;
}

extern void
ba_task_counted_in_warmup_set(ba_task_t *task,
			      ba_real completed_before_warmup,
			      ba_real st_after_warmup)
{
    BA_ASSERT(task);
    BA_ASSERT(task->counted_in_warmup == BA_FALSE);

    task->counted_in_warmup = BA_TRUE;

    task->completed_before_warmup = completed_before_warmup;
    task->st_after_warmup = st_after_warmup;
}

extern void
ba_task_service_delta_time_set(ba_task_t *task,
			       ba_real service_delta_time)
{
    BA_ASSERT(task);
    BA_ASSERT(task->valid_service_delta_time == BA_FALSE);

    task->valid_service_delta_time = BA_TRUE;
    task->service_delta_time = service_delta_time;
}

extern void
ba_task_think_delta_time_set(ba_task_t *task,
			     ba_real think_delta_time)
{
    BA_ASSERT(task);
    BA_ASSERT(task->valid_think_delta_time == BA_FALSE);

    task->valid_think_delta_time = BA_TRUE;
    task->think_delta_time = think_delta_time;
}

extern void
ba_task_will_be_requested_set(ba_task_t *task,
			      ba_bool will_be_requested)
{
    BA_ASSERT(task);
    BA_ASSERT(task->valid_will_be_requested == BA_FALSE);

    task->valid_will_be_requested = BA_TRUE;
    task->will_be_requested = will_be_requested;
}

extern ba_bool
ba_task_running_get(ba_task_t *task)
{
    BA_ASSERT(task);

    return task->running;
}

extern void
ba_task_running_set(ba_task_t *task)
{
    BA_ASSERT(task);
    BA_ASSERT(task->running == BA_FALSE);

    task->running = BA_TRUE;
}

extern void
ba_task_running_unset(ba_task_t *task)
{
    BA_ASSERT(task);
    BA_ASSERT(task->running == BA_TRUE);

    task->running = BA_FALSE;
}

extern ba_bool
ba_task_disclosed_get(ba_task_t *task)
{
    BA_ASSERT(task);

    return task->disclosed;
}

extern ba_bool
ba_task_executed_get(ba_task_t *task)
{
    BA_ASSERT(task);

    return task->executed;
}

extern ba_bool
ba_task_requested_get(ba_task_t *task)
{
    BA_ASSERT(task);

    return task->requested;
}

extern ba_bool
ba_task_finished_get(ba_task_t *task)
{
    BA_ASSERT(task);

    return task->finished;
}

extern ba_bool
ba_task_canceled_get(ba_task_t *task)
{
    BA_ASSERT(task);

    return task->canceled;
}

extern void
ba_task_disclosed_abs_time_set(ba_task_t *task,
			       ba_real disclosed_abs_time)
{
    BA_ASSERT(task);
    BA_ASSERT(task->disclosed == BA_FALSE);

    task->disclosed_abs_time = disclosed_abs_time;
    task->disclosed = BA_TRUE;
}

extern void
ba_task_executed_abs_time_set(ba_task_t *task,
			      ba_real executed_abs_time)
{
    BA_ASSERT(task);
    BA_ASSERT(task->executed == BA_FALSE);

    task->executed_abs_time = executed_abs_time;
    task->executed = BA_TRUE;
}

extern void
ba_task_requested_abs_time_set(ba_task_t *task,
			       ba_real requested_abs_time)
{
    BA_ASSERT(task);
    BA_ASSERT(task->requested == BA_FALSE);

    task->requested_abs_time = requested_abs_time;
    task->requested = BA_TRUE;
}

extern void
ba_task_finished_abs_time_set(ba_task_t *task,
			       ba_real finished_abs_time)
{
    BA_ASSERT(task);
    BA_ASSERT(task->finished == BA_FALSE);

    task->finished_abs_time = finished_abs_time;
    task->finished = BA_TRUE;
}

extern void
ba_task_canceled_abs_time_set(ba_task_t *task,
			     ba_real canceled_abs_time)
{
    BA_ASSERT(task);
    BA_ASSERT(task->canceled == BA_FALSE);

    task->canceled_abs_time = canceled_abs_time;
    task->canceled = BA_TRUE;
}

extern ba_real
ba_task_disclosed_abs_time_get(ba_task_t *task)
{
    BA_ASSERT(task);
    
    BA_ASSERT(task->disclosed);
    return task->disclosed_abs_time;
}

extern ba_real
ba_task_executed_abs_time_get(ba_task_t *task)
{
    BA_ASSERT(task);
    
    BA_ASSERT(task->executed);
    return task->executed_abs_time;
}

extern ba_real
ba_task_requested_abs_time_get(ba_task_t *task)
{
    BA_ASSERT(task);
    
    BA_ASSERT(task->requested);
    return task->requested_abs_time;
}

extern ba_real
ba_task_finished_abs_time_get(ba_task_t *task)
{
    BA_ASSERT(task);
    
    BA_ASSERT(task->finished);
    return task->finished_abs_time;
}

extern ba_real
ba_task_canceled_abs_time_get(ba_task_t *task)
{
    BA_ASSERT(task);
    
    BA_ASSERT(task->canceled);
    return task->canceled_abs_time;
}

extern ba_task_set_t *
ba_task_task_set_get(ba_task_t *task)
{
    BA_ASSERT(task);

    return task->task_set;
}

extern ba_user_t *
ba_task_user_get(ba_task_t *task)
{
    BA_ASSERT(task);
    ba_task_set_t *task_set;
    ba_user_t *user;

    task_set = ba_task_task_set_get(task);
    BA_ASSERT(task_set);
    user = ba_task_set_user_get(task_set);
    BA_ASSERT(user);

    return user;
}

extern ba_real
ba_task_remaining_delta_time_get(ba_task_t *task)
{
    BA_ASSERT(task);
    ba_real remaining_delta_time;

    BA_ASSERT(task->valid_service_delta_time);
    remaining_delta_time = task->service_delta_time -
	task->completed_delta_time;
    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(remaining_delta_time, 0.0);

    return remaining_delta_time;
}

extern void
ba_task_add_completed_delta_time(ba_task_t *task,
				 ba_real add_completed_delta_time)
{
    BA_ASSERT(task);

    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(add_completed_delta_time, 0.0);
    task->completed_delta_time += add_completed_delta_time;
    BA_ASSERT(task->valid_service_delta_time);
    BA_ASSERT_REAL_LESS_THAN_OR_EQUAL(task->completed_delta_time,
				      task->service_delta_time);
}

extern void
ba_task_add_completed_warmup_delta_time(ba_task_t *task,
					ba_real add_completed_warmup_delta_time)
{
    BA_ASSERT(task);

    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(add_completed_warmup_delta_time, 0.0);
    task->completed_warmup_delta_time += add_completed_warmup_delta_time;
    BA_ASSERT(task->valid_service_delta_time);
    BA_ASSERT_REAL_LESS_THAN_OR_EQUAL(task->completed_warmup_delta_time,
				      task->service_delta_time);
    // next assert assumes that ba_task_add_completed_delta_time()
    // called before this for a given slice of this task's execution.
    BA_ASSERT_REAL_LESS_THAN_OR_EQUAL(task->completed_warmup_delta_time,
				      task->completed_delta_time);
}

extern ba_real
ba_task_completed_delta_time_get(ba_task_t *task)
{
    BA_ASSERT(task);
    
    return task->completed_delta_time;
}

extern ba_real
ba_task_completed_warmup_delta_time_get(ba_task_t *task)
{
    BA_ASSERT(task);
    
    return task->completed_warmup_delta_time;
}

extern ba_real
ba_task_service_delta_time_get(ba_task_t *task)
{
    BA_ASSERT(task);
    BA_ASSERT(task->valid_service_delta_time);

    return task->service_delta_time;
}

extern ba_real
ba_task_think_delta_time_get(ba_task_t *task)
{
    BA_ASSERT(task);
    BA_ASSERT(task->valid_think_delta_time);

    return task->think_delta_time;
}

extern ba_real
ba_task_will_be_requested_get(ba_task_t *task)
{
    BA_ASSERT(task);
    BA_ASSERT(task->valid_will_be_requested);

    return task->will_be_requested;
}

extern void
ba_task_print(ba_task_t *task)
{
    BA_ASSERT(task);

    ba_print_info("task: %p\n", task);
    ba_print_info("  id: " BA_ID_FMT "\n", task->id);
    ba_print_info("  valid_service_delta_time: ");
    ba_types_bool_print(task->valid_service_delta_time);
    ba_print_info("\n");
    ba_print_info("  service_delta_time: " BA_REAL_FMT "\n",
		  task->service_delta_time);
    ba_print_info("  valid_think_delta_time: ");
    ba_types_bool_print(task->valid_think_delta_time);
    ba_print_info("\n");
    ba_print_info("  think_delta_time: " BA_REAL_FMT "\n",
		  task->think_delta_time);
    ba_print_info("  valid_will_be_requested: ");
    ba_types_bool_print(task->valid_will_be_requested);
    ba_print_info("\n");
    ba_print_info("  will_be_requested: ");
    ba_types_bool_print(task->will_be_requested);
    ba_print_info("\n");
    ba_print_info("  completed_delta_time: " BA_REAL_FMT "\n",
		  task->completed_delta_time);
    ba_print_info("  completed_warmup_delta_time: " BA_REAL_FMT "\n",
		  task->completed_warmup_delta_time);

    ba_print_info("  running: ");
    ba_types_bool_print(task->running);
    ba_print_info("\n");

    ba_print_info("  disclosed: ");
    ba_types_bool_print(task->disclosed);
    ba_print_info("\n");
    if(task->disclosed) {
	ba_print_info("  disclosed_abs_time: " BA_REAL_FMT "\n",
		      task->disclosed_abs_time);
    }

    ba_print_info("  executed: ");
    ba_types_bool_print(task->executed);
    ba_print_info("\n");
    if(task->executed) {
	ba_print_info("  executed_abs_time: " BA_REAL_FMT "\n",
		      task->executed_abs_time);
    }

    ba_print_info("  requested: ");
    ba_types_bool_print(task->requested);
    ba_print_info("\n");
    if(task->requested) {
	ba_print_info("  requested_abs_time: " BA_REAL_FMT "\n",
		      task->requested_abs_time);
    }

    ba_print_info("  finished: ");
    ba_types_bool_print(task->finished);
    ba_print_info("\n");
    if(task->finished) {
	ba_print_info("  finished_abs_time: " BA_REAL_FMT "\n",
		      task->finished_abs_time);
    }

    ba_print_info("  canceled: ");
    ba_types_bool_print(task->canceled);
    ba_print_info("\n");
    if(task->canceled) {
	ba_print_info("  canceled_abs_time: " BA_REAL_FMT "\n",
		      task->canceled_abs_time);
    }

    ba_print_info("  task_set: %p\n", task->task_set);
}

extern void
ba_task_print_from_list(void *task_in,
			void *unused)
{
    BA_ASSERT(task_in);
    ba_task_t *task;
    
    task = (ba_task_t *)task_in;

    ba_task_print(task);
}

extern ba_task_t *
ba_task_clone_new_from_old(ba_task_t *task_old,
			   ba_task_set_t *task_set)
{
    BA_ASSERT(task_old && task_set);
    ba_task_t *task_new;    

    task_new = ba_mem_malloc(sizeof(ba_task_t));

    task_new->id = task_old->id;

    task_new->valid_service_delta_time = task_old->valid_service_delta_time;
    task_new->service_delta_time = task_old->service_delta_time;
    task_new->valid_think_delta_time = task_old->valid_think_delta_time;
    task_new->think_delta_time = task_old->think_delta_time;
    task_new->valid_will_be_requested = task_old->valid_will_be_requested;
    task_new->will_be_requested = task_old->will_be_requested;
    task_new->completed_delta_time = task_old->completed_delta_time;
    task_new->completed_warmup_delta_time =
	task_old->completed_warmup_delta_time;

    task_new->running = task_old->running;

    task_new->disclosed = task_old->disclosed;
    task_new->disclosed_abs_time = task_old->disclosed_abs_time;
    task_new->executed = task_old->executed;
    task_new->executed_abs_time = task_old->executed_abs_time;
    task_new->requested = task_old->requested;
    task_new->requested_abs_time = task_old->requested_abs_time;
    task_new->finished = task_old->finished;
    task_new->finished_abs_time = task_old->finished_abs_time;
    task_new->canceled = task_old->canceled;
    task_new->canceled_abs_time = task_old->canceled_abs_time;

    task_new->task_set = task_set;

    return task_new;
}

extern ba_task_t *
ba_task_new(ba_id_t id,
	    ba_task_set_t *task_set)
{
    BA_ASSERT(task_set);
    ba_task_t *task;

    task = ba_mem_malloc(sizeof(ba_task_t));

    task->id = id;

    task->running = BA_FALSE;

    task->valid_service_delta_time = BA_FALSE;
    task->service_delta_time = 0.0;
    task->valid_think_delta_time = BA_FALSE;
    task->think_delta_time = 0.0;
    task->valid_will_be_requested = BA_FALSE;
    task->will_be_requested = BA_FALSE;
    task->counted_in_warmup = BA_FALSE;
    task->completed_before_warmup = 0.0;
    task->st_after_warmup = 0.0;
    task->completed_delta_time = 0.0;
    task->completed_warmup_delta_time = 0.0;

    task->disclosed = BA_FALSE;
    task->disclosed_abs_time = 0.0;
    task->executed = BA_FALSE;
    task->executed_abs_time = 0.0;
    task->requested = BA_FALSE;
    task->requested_abs_time = 0.0;
    task->finished = BA_FALSE;
    task->finished_abs_time = 0.0;
    task->canceled = BA_FALSE;
    task->canceled_abs_time = 0.0;

    task->task_set = task_set;

    return task;
}

extern void
ba_task_delete(ba_task_t *task)
{
    BA_ASSERT(task);

    BA_ASSERT(task->running == BA_FALSE);

    ba_mem_free(task);
}

extern void
ba_task_delete_from_list(void *task_in,
			 void *unused)
{
    BA_ASSERT(task_in);
    ba_task_t *task;

    task = (ba_task_t *)task_in;

    ba_task_delete(task);
}
