#include <ba_config.h>
#include <ba_sim.h>

struct ba_task_set_s {
    ba_id_t id;

    GSList *tasks;

    ba_bool valid_activated_abs_time;
    ba_real activated_abs_time;
    ba_bool valid_deactivated_abs_time;
    ba_real deactivated_abs_time;

    ba_bool valid_last_task_executed_abs_time;
    ba_real last_task_executed_abs_time;

    ba_user_t *user;
};

extern void
ba_task_set_report(ba_task_set_t *task_set)
{
    BA_ASSERT(task_set);
    ba_real ready_end;
    ba_bool found;
    GSList *tasks;

    // set task_set_alive_start_end
    BA_ASSERT(task_set->valid_activated_abs_time &&
	      task_set->valid_deactivated_abs_time);
    ba_report_task_set_alive_start_end(task_set->activated_abs_time,
				       task_set->deactivated_abs_time);

    // set task_set_ready_tasks_start_end
    found = BA_FALSE;
    ready_end = 0.0; // warning elimination
    for(tasks = task_set->tasks; tasks; tasks = g_slist_next(tasks)) {
	ba_task_t *task;

	task = (ba_task_t *)tasks->data;
	BA_ASSERT(task);
	if(ba_task_executed_get(task) == BA_FALSE &&
	   (ba_task_requested_get(task) || ba_task_disclosed_get(task))) {
	    found = BA_TRUE;
	    ready_end = task_set->deactivated_abs_time;
	    break;
	}
    }
    if(found == BA_FALSE) {
	BA_ASSERT(task_set->valid_last_task_executed_abs_time);
	ready_end = task_set->last_task_executed_abs_time;
    }
    ba_report_task_set_ready_tasks_start_end(task_set->activated_abs_time,
					     ready_end);
}

extern void
ba_task_set_activated_abs_time_set(ba_task_set_t *task_set,
				   ba_real activated_abs_time)
{
    BA_ASSERT(task_set);

    BA_ASSERT(task_set->valid_activated_abs_time == BA_FALSE);
    task_set->activated_abs_time = activated_abs_time;
    task_set->valid_activated_abs_time = BA_TRUE;
}

extern ba_real
ba_task_set_activated_abs_time_get(ba_task_set_t *task_set)
{
    BA_ASSERT(task_set);

    BA_ASSERT(task_set->valid_activated_abs_time);

    return task_set->activated_abs_time;
}

extern void
ba_task_set_deactivated_abs_time_set(ba_task_set_t *task_set,
				     ba_real deactivated_abs_time)
{
    BA_ASSERT(task_set);

    BA_ASSERT(task_set->valid_deactivated_abs_time == BA_FALSE);
    task_set->deactivated_abs_time = deactivated_abs_time;
    task_set->valid_deactivated_abs_time = BA_TRUE;
}

extern void
ba_task_set_last_task_executed_abs_time_set(ba_task_set_t *task_set,
				    ba_real last_task_executed_abs_time)
{
    BA_ASSERT(task_set);

    task_set->valid_last_task_executed_abs_time = BA_TRUE;
    task_set->last_task_executed_abs_time = last_task_executed_abs_time;
}

extern ba_real
ba_task_set_last_task_executed_abs_time_get(ba_task_set_t *task_set)
{
    BA_ASSERT(task_set);

    BA_ASSERT(task_set->valid_last_task_executed_abs_time);
    return task_set->last_task_executed_abs_time;
}

extern ba_id_t
ba_task_set_id_get(ba_task_set_t *task_set)
{
    BA_ASSERT(task_set);

    return task_set->id;
}

extern ba_user_t *
ba_task_set_user_get(ba_task_set_t *task_set)
{
    BA_ASSERT(task_set);
    
    return task_set->user;
}

extern ba_task_t *
ba_task_set_return_unexecuted_task(ba_task_set_t *task_set)
{
    BA_ASSERT(task_set);
    GSList *tasks;

    for(tasks = task_set->tasks; tasks; tasks = g_slist_next(tasks)) {
	ba_task_t *task;
	ba_bool executed;

	task = (ba_task_t *)tasks->data;
	BA_ASSERT(task);
	executed = ba_task_executed_get(task);
	if(executed == BA_FALSE) {
	    return task;
	}
    }
    return 0;
}

extern ba_task_t *
ba_task_set_return_next_unrequested_task(ba_task_set_t *task_set)
{
    BA_ASSERT(task_set);
    GSList *tasks;

    for(tasks = task_set->tasks; tasks; tasks = g_slist_next(tasks)) {
	ba_task_t *task;
	ba_bool requested;

	task = (ba_task_t *)tasks->data;
	BA_ASSERT(task);
	requested = ba_task_requested_get(task);
	if(requested == BA_FALSE) {
	    return task;
	}
    }
    return 0;
}

extern void
ba_task_set_append_task(ba_task_set_t *task_set,
			ba_task_t *task)
{
    BA_ASSERT(task_set);
    BA_ASSERT(task);

    task_set->tasks = g_slist_append(task_set->tasks, task);
}

extern void
ba_task_set_disclose_all(ba_task_set_t *task_set)
{
    BA_ASSERT(task_set);
    ba_user_t *user;
    GSList *tasks;

    user = task_set->user;

    for(tasks = task_set->tasks; tasks; tasks = g_slist_next(tasks)) {
	ba_task_t *task;

	task = (ba_task_t *)tasks->data;
	BA_ASSERT(task);
	if(ba_user_can_req_or_dis(user, BA_FALSE) == BA_FALSE) {
	    return;
	}
	ba_user_add_disclosed_task(user, task);
    }
}

extern ba_bool
ba_task_set_is_empty(ba_task_set_t *task_set)
{
    BA_ASSERT(task_set);

    if(task_set->tasks == 0) {
	return BA_TRUE;
    } else {
	return BA_FALSE;
    }
}

extern void
ba_task_set_print(ba_task_set_t *task_set)
{
    BA_ASSERT(task_set);

    ba_print_info("task_set: %p\n", task_set);
    ba_print_info("  id: " BA_ID_FMT "\n", task_set->id);

    ba_print_info("  valid_activated_abs_time: ");
    ba_types_bool_print(task_set->valid_activated_abs_time);
    ba_print_info("\n");
    if(task_set->valid_activated_abs_time) {
	ba_print_info("  activated_abs_time: " BA_REAL_FMT "\n",
		      task_set->activated_abs_time);
    }

    ba_print_info("  valid_deactivated_abs_time: ");
    ba_types_bool_print(task_set->valid_deactivated_abs_time);
    ba_print_info("\n");
    if(task_set->valid_deactivated_abs_time) {
	ba_print_info("  deactivated_abs_time: " BA_REAL_FMT "\n",
		      task_set->deactivated_abs_time);
    }

    ba_print_info("  valid_last_task_executed_abs_time: ");
    ba_types_bool_print(task_set->valid_last_task_executed_abs_time);
    ba_print_info("\n");
    if(task_set->valid_last_task_executed_abs_time) {
	ba_print_info("  last_task_executed_abs_time: " BA_REAL_FMT "\n",
		      task_set->last_task_executed_abs_time);
    }

    ba_print_info("  user: %p\n", task_set->user);
    ba_print_info("  (skipping the printing of tasks in this set)\n");
    //    g_slist_foreach(task_set->tasks, ba_task_print_from_list, 0);
}

extern void
ba_task_set_print_from_list(void *task_set_in,
			    void *unused)
{
    BA_ASSERT(task_set_in);
    ba_task_set_t *task_set;
    
    task_set = (ba_task_set_t *)task_set_in;

    ba_task_set_print(task_set);
}

extern ba_task_set_t *
ba_task_set_new(ba_id_t id,
		ba_user_t *user)
{
    BA_ASSERT(user);
    ba_task_set_t *task_set;

    task_set = ba_mem_malloc(sizeof(ba_task_set_t));

    task_set->id = id;
    task_set->tasks = 0;

    task_set->valid_activated_abs_time = BA_FALSE;
    task_set->activated_abs_time = 0.0;
    task_set->valid_deactivated_abs_time = BA_FALSE;
    task_set->deactivated_abs_time = 0.0;

    task_set->valid_last_task_executed_abs_time = BA_FALSE;
    task_set->last_task_executed_abs_time = 0.0;

    task_set->user = user;

    return task_set;
}

extern void
ba_task_set_delete(ba_task_set_t *task_set)
{
    BA_ASSERT(task_set);

    g_slist_foreach(task_set->tasks, ba_task_delete_from_list, 0);
    g_slist_free(task_set->tasks);

    ba_mem_free(task_set);
}

extern void
ba_task_set_delete_from_list(void *task_set_in,
			     void *unused)
{
    BA_ASSERT(task_set_in);
    ba_task_set_t *task_set;

    task_set = (ba_task_set_t *)task_set_in;

    ba_task_set_delete(task_set);
}
