#include <ba_config.h>
#include <ba_sim.h>

struct ba_time_period_s {
    ba_real start_abs_time;
    ba_real end_abs_time;
};

extern ba_bool
ba_time_period_is_equal(ba_time_period_t *a_time_period,
			ba_time_period_t *b_time_period)
{
    BA_ASSERT(a_time_period && b_time_period);

    if(ba_types_real_equal(a_time_period->start_abs_time,
			   b_time_period->start_abs_time) &&
       ba_types_real_equal(a_time_period->end_abs_time,
			   b_time_period->end_abs_time)) {
	return BA_TRUE;
    } else {
	return BA_FALSE;
    }
}

extern ba_real
ba_time_period_start_abs_time_get(ba_time_period_t *time_period)
{
    BA_ASSERT(time_period);

    return time_period->start_abs_time;
}

extern ba_real
ba_time_period_end_abs_time_get(ba_time_period_t *time_period)
{
    BA_ASSERT(time_period);

    return time_period->end_abs_time;
}

extern ba_real
ba_time_period_delta_time(ba_time_period_t *time_period)
{
    BA_ASSERT(time_period);

    return time_period->end_abs_time - time_period->start_abs_time;
}

extern void
ba_time_period_print(ba_time_period_t *time_period)
{
    BA_ASSERT(time_period);

    ba_print_info("Time period %p: [" BA_REAL_FMT ", " BA_REAL_FMT "]\n",
		  time_period, time_period->start_abs_time,
		  time_period->end_abs_time);
}

extern void
ba_time_period_print_from_list(void *time_period_in,
			       void *unused)
{
    BA_ASSERT(time_period_in);
    ba_time_period_t *time_period;
    
    time_period = (ba_time_period_t *)time_period_in;
    
    ba_time_period_print(time_period);
}

extern ba_time_period_t *
ba_time_period_copy(ba_time_period_t *time_period_src)
{
    BA_ASSERT(time_period_src);
    ba_time_period_t *time_period_dst;

    time_period_dst = ba_time_period_new(time_period_src->start_abs_time,
					 time_period_src->end_abs_time);

    return time_period_dst;
}

extern ba_time_period_t *
ba_time_period_new(ba_real start_abs_time,
		   ba_real end_abs_time)
{
    // we allow a 0-length period...
    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(end_abs_time, start_abs_time);
    ba_time_period_t *time_period;

    time_period = ba_mem_malloc(sizeof(ba_time_period_t));

    time_period->start_abs_time = start_abs_time;
    time_period->end_abs_time = end_abs_time;

    return time_period;
}

extern void
ba_time_period_delete(ba_time_period_t *time_period)
{
    BA_ASSERT(time_period);

    ba_mem_free(time_period);
}

extern void
ba_time_period_delete_from_list(void *time_period_in,
				void *unused)
{
    BA_ASSERT(time_period_in);
    ba_time_period_t *time_period;
    
    time_period = (ba_time_period_t *)time_period_in;
    
    ba_time_period_delete(time_period);
}
