#include <ba_config.h>
#include <ba_sim.h>

struct ba_time_periods_s {
    // what else could go here?  an identifying name?
    GSList *periods;
};

static void swap_periods(ba_time_periods_t *a_time_periods,
			 ba_time_periods_t *b_time_periods);
static void simplify_periods(ba_time_periods_t *time_periods);

static ba_time_periods_t *union_new_opt_simplify(ba_time_periods_t
						 *a_time_periods,
						 ba_time_periods_t
						 *b_time_periods,
						 ba_bool simplify);
static void union_helper(ba_time_periods_t *union_periods,
			 ba_real start_abs_time,
			 ba_real end_abs_time);

static ba_time_periods_t *intersection_new_opt_simplify(ba_time_periods_t
							*a_time_periods,
							ba_time_periods_t
							*b_time_periods,
							ba_bool simplify);

static ba_time_periods_t *difference_new_opt_simplify(ba_time_periods_t
						      *a_time_periods,
						      ba_time_periods_t
						      *b_time_periods,
						      ba_bool simplify);

static ba_time_periods_t *complement_new_opt_simplify(ba_time_periods_t
						      *original,
						      ba_time_period_t
						      *universe,
						      ba_bool simplify);

static void append_period_opt_copy_opt_simplify(ba_time_periods_t
						*time_periods,
						ba_time_period_t *time_period,
						ba_bool copy,
						ba_bool simplify);
static void append_start_end(ba_time_periods_t *time_periods,
			     ba_real start_abs_time,
			     ba_real end_abs_time);

// static utils //////////////////////////////////////////////////////

// be wary if i add stuff to ba_periods_t
static void
swap_periods(ba_time_periods_t *a_time_periods,
	     ba_time_periods_t *b_time_periods)
{
    BA_ASSERT(a_time_periods && b_time_periods);
    GSList *tmp;

    tmp = a_time_periods->periods;
    a_time_periods->periods = b_time_periods->periods;
    b_time_periods->periods = tmp;
}

static void
simplify_periods(ba_time_periods_t *time_periods)
{
    BA_ASSERT(time_periods);
    ba_time_period_t *time_period, *next_time_period, *new_time_period;
    GSList *elem, *next_elem, *next_next_elem;
    ba_real time_period_start_abs_time, time_period_end_abs_time;
    ba_real next_time_period_start_abs_time, next_time_period_end_abs_time;

    if(ba_time_periods_is_empty(time_periods)) {
	return;
    }
    
 repeat:
    elem = time_periods->periods;
    if(g_slist_next(elem) == 0) {
	return;
    }
    next_elem = g_slist_next(elem);
	
    while(next_elem) {
	time_period = (ba_time_period_t *)elem->data;
	next_time_period = (ba_time_period_t *)next_elem->data;
	time_period_start_abs_time =
	    ba_time_period_start_abs_time_get(time_period);
	time_period_end_abs_time =
	    ba_time_period_end_abs_time_get(time_period);
	next_time_period_start_abs_time =
	    ba_time_period_start_abs_time_get(next_time_period);
	next_time_period_end_abs_time =
	    ba_time_period_end_abs_time_get(next_time_period);

	if(ba_types_real_equal(time_period_end_abs_time,
			       next_time_period_start_abs_time)) {
	    next_next_elem = g_slist_next(next_elem);
	    time_periods->periods = g_slist_remove(time_periods->periods,
						   time_period);
	    ba_time_period_delete(time_period);
	    
	    time_periods->periods = g_slist_remove(time_periods->periods,
						   next_time_period);
	    ba_time_period_delete(next_time_period);
	    
	    new_time_period =
		ba_time_period_new(time_period_start_abs_time,
				   next_time_period_end_abs_time);
	    time_periods->periods =
		g_slist_insert_before(time_periods->periods, next_next_elem,
				      new_time_period);
	    goto repeat;
	}

	elem = g_slist_next(elem);
	next_elem = g_slist_next(next_elem);
    }
}

// extern utils //////////////////////////////////////////////////////

extern void
ba_time_periods_nullify(ba_time_periods_t *time_periods)
{
    BA_ASSERT(time_periods);

    g_slist_foreach(time_periods->periods, ba_time_period_delete_from_list, 0);
    g_slist_free(time_periods->periods);

    time_periods->periods = 0;
}

extern ba_bool
ba_time_periods_is_equal(ba_time_periods_t *a_time_periods,
			 ba_time_periods_t *b_time_periods)
{
    BA_ASSERT(a_time_periods && b_time_periods);
    GSList *a_elem, *b_elem;
    ba_time_period_t *a_time_period, *b_time_period;

    for(a_elem = a_time_periods->periods, b_elem = b_time_periods->periods;
	a_elem && b_elem;
	a_elem = g_slist_next(a_elem), b_elem = g_slist_next(b_elem)) {
	a_time_period = (ba_time_period_t *)a_elem->data;
	b_time_period = (ba_time_period_t *)b_elem->data;

	if(ba_time_period_is_equal(a_time_period, b_time_period) == BA_FALSE) {
	    return BA_FALSE;
	}
    }
    if(a_elem || b_elem) {
	return BA_FALSE;
    }

    return BA_TRUE;
}

extern ba_bool
ba_time_periods_is_equal_period(ba_time_periods_t *a_time_periods,
				ba_time_period_t *b_time_period)
{
    BA_ASSERT(a_time_periods && b_time_period);
    ba_time_periods_t *b_time_periods;
    ba_bool equal;

    b_time_periods = ba_time_periods_new();

    ba_time_periods_append_period(b_time_periods, b_time_period);

    equal = ba_time_periods_is_equal(a_time_periods, b_time_periods);

    ba_time_periods_delete(b_time_periods);
    return equal;
}

extern ba_bool
ba_time_periods_is_equal_start_end(ba_time_periods_t *a_time_periods,
				   ba_real start_abs_time,
				   ba_real end_abs_time)
{
    BA_ASSERT(a_time_periods);
    ba_time_periods_t *b_time_periods;
    ba_bool equal;

    b_time_periods = ba_time_periods_new();

    ba_time_periods_append_start_end(b_time_periods, start_abs_time,
				     end_abs_time);

    equal = ba_time_periods_is_equal(a_time_periods, b_time_periods);

    ba_time_periods_delete(b_time_periods);
    return equal;
}

extern ba_bool
ba_time_periods_is_empty(ba_time_periods_t *time_periods)
{
    BA_ASSERT(time_periods);

    if(time_periods->periods == 0) {
	return BA_TRUE;
    } else {
	return BA_FALSE;
    }
}

extern ba_real
ba_time_periods_total_delta_time(ba_time_periods_t *time_periods)
{
    BA_ASSERT(time_periods);
    ba_time_period_t *time_period;
    GSList *elem;
    ba_real total_delta_time;

    total_delta_time = 0.0;
    for(elem = time_periods->periods; elem; elem = g_slist_next(elem)) {
	time_period = (ba_time_period_t *)elem->data;

	total_delta_time += ba_time_period_end_abs_time_get(time_period) -
	    ba_time_period_start_abs_time_get(time_period);
    }
    
    return total_delta_time;
}

extern void
ba_time_periods_print(ba_time_periods_t *time_periods)
{
    BA_ASSERT(time_periods);

    ba_print_info("Time periods %p (%p):\n", time_periods,
		  time_periods->periods);
    g_slist_foreach(time_periods->periods, ba_time_period_print_from_list, 0);
}

extern ba_real
ba_time_periods_first_start_abs_time(ba_time_periods_t *time_periods)
{
    BA_ASSERT(time_periods);
    ba_time_period_t *time_period;
    GSList *elem;

    BA_ASSERT(ba_time_periods_is_empty(time_periods) == BA_FALSE);

    elem = time_periods->periods;
    time_period = (ba_time_period_t *)elem->data;
    BA_ASSERT(time_period);

    return ba_time_period_start_abs_time_get(time_period);
}

extern ba_real
ba_time_periods_last_end_abs_time(ba_time_periods_t *time_periods)
{
    BA_ASSERT(time_periods);
    ba_time_period_t *time_period;
    GSList *elem;
    
    BA_ASSERT(ba_time_periods_is_empty(time_periods) == BA_FALSE);

    elem = g_slist_last(time_periods->periods);
    time_period = (ba_time_period_t *)elem->data;
    BA_ASSERT(time_period);

    return ba_time_period_end_abs_time_get(time_period);
}

// union /////////////////////////////////////////////////////////////

static ba_time_periods_t *
union_new_opt_simplify(ba_time_periods_t *a_time_periods,
		       ba_time_periods_t *b_time_periods,
		       ba_bool simplify)
{
    BA_ASSERT(a_time_periods && b_time_periods);
    // so as to not conflict with keyword.
    ba_time_periods_t *union_periods;
    // compiler is too dumb to see that the following are def'd before
    // use:
    ba_real a_start_abs_time = 0.0, a_end_abs_time = 0.0;
    ba_real b_start_abs_time = 0.0, b_end_abs_time = 0.0;
    ba_time_period_t *a_time_period, *b_time_period;
    GSList *a_elem, *b_elem;
    ba_real min_start_abs_time, max_end_abs_time;

    union_periods = ba_time_periods_new();

    a_elem = a_time_periods->periods;
    b_elem = b_time_periods->periods;
    while(a_elem || b_elem) {
	if(a_elem) {
	    a_time_period = (ba_time_period_t *)a_elem->data;
	    BA_ASSERT(a_time_period);
	    a_start_abs_time =
		ba_time_period_start_abs_time_get(a_time_period);
	    a_end_abs_time = ba_time_period_end_abs_time_get(a_time_period);
	}
	if(b_elem) {
	    b_time_period = (ba_time_period_t *)b_elem->data;
	    BA_ASSERT(b_time_period);
	    b_start_abs_time =
		ba_time_period_start_abs_time_get(b_time_period);
	    b_end_abs_time = ba_time_period_end_abs_time_get(b_time_period);
	}
	
	if(a_elem && b_elem) {
	    if(a_end_abs_time < b_start_abs_time) {
		union_helper(union_periods, a_start_abs_time, a_end_abs_time);
		a_elem = g_slist_next(a_elem);
		continue;	    
	    } else if(b_end_abs_time < a_start_abs_time) {
		union_helper(union_periods, b_start_abs_time, b_end_abs_time);
		b_elem = g_slist_next(b_elem);
		continue;
	    } else {
		min_start_abs_time = BA_REAL_MIN(a_start_abs_time,
						 b_start_abs_time);
		max_end_abs_time = BA_REAL_MAX(a_end_abs_time, b_end_abs_time);
		union_helper(union_periods, min_start_abs_time,
			     max_end_abs_time);
		a_elem = g_slist_next(a_elem);
		b_elem = g_slist_next(b_elem);
		continue;
	    }
	}
	if(a_elem) {
	    union_helper(union_periods, a_start_abs_time, a_end_abs_time);
	    a_elem = g_slist_next(a_elem);
	} else if(b_elem) {
	    union_helper(union_periods, b_start_abs_time, b_end_abs_time);
	    b_elem = g_slist_next(b_elem);
	} else {
	    BA_PANIC();
	}
    }

    if(simplify) {
	simplify_periods(union_periods);
    }
    return union_periods;
}

static void
union_helper(ba_time_periods_t *union_periods,
	     ba_real start_abs_time,
	     ba_real end_abs_time)
{
    ba_real union_last_end_abs_time;
    ba_real max_start_abs_time;
    
    if(ba_time_periods_is_empty(union_periods) == BA_FALSE) {
	union_last_end_abs_time =
	    ba_time_periods_last_end_abs_time(union_periods);
	max_start_abs_time = BA_REAL_MAX(union_last_end_abs_time,
					 start_abs_time);
    } else {
	max_start_abs_time = start_abs_time;
    }

    if(end_abs_time > max_start_abs_time) {
	append_start_end(union_periods, max_start_abs_time, end_abs_time);
    }
}

extern void
ba_time_periods_union(ba_time_periods_t *a_time_periods,
		      ba_time_periods_t *b_time_periods,
		      ba_time_periods_t *result_time_periods)
{
    BA_ASSERT(a_time_periods && b_time_periods && result_time_periods);
    ba_time_periods_t *tmp_time_periods;

    tmp_time_periods = union_new_opt_simplify(a_time_periods, b_time_periods,
					      BA_TRUE);

    swap_periods(tmp_time_periods, result_time_periods);
    ba_time_periods_delete(tmp_time_periods);
}

extern void
ba_time_periods_union_period(ba_time_periods_t *a_time_periods,
			     ba_time_period_t *b_time_period,
			     ba_time_periods_t *result_time_periods)
{
    BA_ASSERT(a_time_periods && b_time_period && result_time_periods);
    ba_time_periods_t *b_time_periods;

    b_time_periods = ba_time_periods_new();
    ba_time_periods_append_period(b_time_periods, b_time_period);

    ba_time_periods_union(a_time_periods, b_time_periods, result_time_periods);

    ba_time_periods_delete(b_time_periods);
}

extern void
ba_time_periods_union_start_end(ba_time_periods_t *a_time_periods,
				ba_real start_abs_time,
				ba_real end_abs_time,
				ba_time_periods_t *result_time_periods)
{
    BA_ASSERT(a_time_periods && result_time_periods);
    ba_time_periods_t *b_time_periods;

    b_time_periods = ba_time_periods_new();
    ba_time_periods_append_start_end(b_time_periods, start_abs_time,
				     end_abs_time);

    ba_time_periods_union(a_time_periods, b_time_periods, result_time_periods);

    ba_time_periods_delete(b_time_periods);
}

// intersection //////////////////////////////////////////////////////

static ba_time_periods_t *
intersection_new_opt_simplify(ba_time_periods_t *a_time_periods,
			      ba_time_periods_t *b_time_periods,
			      ba_bool simplify)
{
    BA_ASSERT(a_time_periods && b_time_periods);
    ba_time_periods_t *intersection;
    ba_real a_start_abs_time, a_end_abs_time;
    ba_real b_start_abs_time, b_end_abs_time;
    ba_time_period_t *a_time_period, *b_time_period;
    GSList *a_elem, *b_elem;
    ba_real max_start_abs_time, min_end_abs_time;

    intersection = ba_time_periods_new();

    a_elem = a_time_periods->periods;
    b_elem = b_time_periods->periods;
    while(a_elem && b_elem) {
	a_time_period = (ba_time_period_t *)a_elem->data;
	BA_ASSERT(a_time_period);
	b_time_period = (ba_time_period_t *)b_elem->data;
	BA_ASSERT(b_time_period);
	a_start_abs_time = ba_time_period_start_abs_time_get(a_time_period);
	a_end_abs_time = ba_time_period_end_abs_time_get(a_time_period);
	b_start_abs_time = ba_time_period_start_abs_time_get(b_time_period);
	b_end_abs_time = ba_time_period_end_abs_time_get(b_time_period);
	
	max_start_abs_time = BA_REAL_MAX(a_start_abs_time, b_start_abs_time);
	min_end_abs_time = BA_REAL_MIN(a_end_abs_time, b_end_abs_time);
	if(min_end_abs_time > max_start_abs_time) {
	    append_start_end(intersection, max_start_abs_time,
			     min_end_abs_time);
	}

	if(a_end_abs_time > b_end_abs_time) {
	    b_elem = g_slist_next(b_elem);
	} else {
	    a_elem = g_slist_next(a_elem);
	}
    }
    // if one or the other remains, that's ok; intersection ignores them

    if(simplify) {
	simplify_periods(intersection);
    }
    return intersection;
}

extern void
ba_time_periods_intersection(ba_time_periods_t *a_time_periods,
			     ba_time_periods_t *b_time_periods,
			     ba_time_periods_t *result_time_periods)
{
    BA_ASSERT(a_time_periods && b_time_periods && result_time_periods);
    ba_time_periods_t *tmp_time_periods;

    tmp_time_periods = intersection_new_opt_simplify(a_time_periods,
						     b_time_periods, BA_TRUE);

    swap_periods(tmp_time_periods, result_time_periods);
    ba_time_periods_delete(tmp_time_periods);
}

extern void
ba_time_periods_intersection_period(ba_time_periods_t *a_time_periods,
				    ba_time_period_t *b_time_period,
				    ba_time_periods_t *result_time_periods)
{
    BA_ASSERT(a_time_periods && b_time_period && result_time_periods);
    ba_time_periods_t *b_time_periods;
    
    b_time_periods = ba_time_periods_new();
    ba_time_periods_append_period(b_time_periods, b_time_period);
    
    ba_time_periods_intersection(a_time_periods, b_time_periods,
				 result_time_periods);
    
    ba_time_periods_delete(b_time_periods);
}

extern void
ba_time_periods_intersection_start_end(ba_time_periods_t *a_time_periods,
				       ba_real start_abs_time,
				       ba_real end_abs_time,
				       ba_time_periods_t *result_time_periods)
{
    BA_ASSERT(a_time_periods && result_time_periods);
    ba_time_periods_t *b_time_periods;
    
    b_time_periods = ba_time_periods_new();
    ba_time_periods_append_start_end(b_time_periods, start_abs_time,
				     end_abs_time);

    ba_time_periods_intersection(a_time_periods, b_time_periods,
				 result_time_periods);
    
    ba_time_periods_delete(b_time_periods);
}

// difference ////////////////////////////////////////////////////////

static ba_time_periods_t *
difference_new_opt_simplify(ba_time_periods_t *a_time_periods,
			    ba_time_periods_t *b_time_periods,
			    ba_bool simplify)
{
    BA_ASSERT(a_time_periods && b_time_periods);
    ba_time_periods_t *difference;
    ba_time_periods_t *complement;
    ba_time_period_t *universe;
    ba_real a_first_start_abs_time, a_last_end_abs_time;
    ba_real b_first_start_abs_time, b_last_end_abs_time;
    ba_real min_start_abs_time, max_end_abs_time;
    
    if(ba_time_periods_is_empty(a_time_periods)) {
	difference = ba_time_periods_new(); // empty
	return difference;
    }
    if(ba_time_periods_is_empty(b_time_periods)) {
	difference = ba_time_periods_copy(a_time_periods);
	return difference;
    }
    
    a_first_start_abs_time =
	ba_time_periods_first_start_abs_time(a_time_periods);
    b_first_start_abs_time =
	ba_time_periods_first_start_abs_time(b_time_periods);
    a_last_end_abs_time = ba_time_periods_last_end_abs_time(a_time_periods);
    b_last_end_abs_time = ba_time_periods_last_end_abs_time(b_time_periods);

    min_start_abs_time = BA_REAL_MIN(a_first_start_abs_time,
				     b_first_start_abs_time);
    max_end_abs_time = BA_REAL_MAX(a_last_end_abs_time, b_last_end_abs_time);

    universe = ba_time_period_new(min_start_abs_time, max_end_abs_time);

    complement = complement_new_opt_simplify(b_time_periods, universe,
					     BA_FALSE);
    difference = intersection_new_opt_simplify(a_time_periods, complement,
					       BA_FALSE);

    ba_time_period_delete(universe);
    ba_time_periods_delete(complement);

    if(simplify) {
	simplify_periods(difference);
    }

    return difference;
}

extern void
ba_time_periods_difference(ba_time_periods_t *a_time_periods,
			   ba_time_periods_t *b_time_periods,
			   ba_time_periods_t *result_time_periods)
{
    BA_ASSERT(a_time_periods && b_time_periods && result_time_periods);
    ba_time_periods_t *tmp_time_periods;

    tmp_time_periods = difference_new_opt_simplify(a_time_periods,
						   b_time_periods, BA_TRUE);

    swap_periods(tmp_time_periods, result_time_periods);
    ba_time_periods_delete(tmp_time_periods);
}

extern void
ba_time_periods_difference_period(ba_time_periods_t *a_time_periods,
				  ba_time_period_t *b_time_period,
				  ba_time_periods_t *result_time_periods)
{
    BA_ASSERT(a_time_periods && b_time_period && result_time_periods);
    ba_time_periods_t *b_time_periods;
    
    b_time_periods = ba_time_periods_new();
    ba_time_periods_append_period(b_time_periods, b_time_period);
    
    ba_time_periods_difference(a_time_periods, b_time_periods,
			       result_time_periods);

    ba_time_periods_delete(b_time_periods);
}

extern void
ba_time_periods_difference_period_r(ba_time_period_t *a_time_period,
				    ba_time_periods_t *b_time_periods,
				    ba_time_periods_t *result_time_periods)
{
    BA_ASSERT(a_time_period && b_time_periods && result_time_periods);
    ba_time_periods_t *a_time_periods;
    
    a_time_periods = ba_time_periods_new();
    ba_time_periods_append_period(a_time_periods, a_time_period);
    
    ba_time_periods_difference(a_time_periods, b_time_periods,
			       result_time_periods);
    
    ba_time_periods_delete(a_time_periods);
}

extern void
ba_time_periods_difference_start_end(ba_time_periods_t *a_time_periods,
				     ba_real start_abs_time,
				     ba_real end_abs_time,
				     ba_time_periods_t *result_time_periods)
{
    BA_ASSERT(a_time_periods && result_time_periods);
    ba_time_periods_t *b_time_periods;
    
    b_time_periods = ba_time_periods_new();
    ba_time_periods_append_start_end(b_time_periods, start_abs_time,
				     end_abs_time);
    
    ba_time_periods_difference(a_time_periods, b_time_periods,
			       result_time_periods);

    ba_time_periods_delete(b_time_periods);
}

extern void
ba_time_periods_difference_start_end_r(ba_real start_abs_time,
				       ba_real end_abs_time,
				       ba_time_periods_t *b_time_periods,
				       ba_time_periods_t *result_time_periods)
{
    BA_ASSERT(b_time_periods && result_time_periods);
    ba_time_periods_t *a_time_periods;

    a_time_periods = ba_time_periods_new();
    ba_time_periods_append_start_end(a_time_periods, start_abs_time,
				     end_abs_time);
    
    ba_time_periods_difference(a_time_periods, b_time_periods,
			       result_time_periods);

    ba_time_periods_delete(a_time_periods);
}

// complement ////////////////////////////////////////////////////////

static ba_time_periods_t *
complement_new_opt_simplify(ba_time_periods_t *original,
			    ba_time_period_t *universe,
			    ba_bool simplify)
{
    BA_ASSERT(original && universe);
    ba_time_periods_t *complement;
    ba_real start_abs_time, end_abs_time;
    ba_real universe_start_abs_time, universe_end_abs_time;
    GSList *elem;
    ba_time_period_t *time_period;
    ba_real current_abs_time;

    universe_start_abs_time = ba_time_period_start_abs_time_get(universe);
    universe_end_abs_time = ba_time_period_end_abs_time_get(universe);

    if(ba_time_periods_is_empty(original) == BA_FALSE) {
	// make sure original doesn't start before universe
	start_abs_time = ba_time_periods_first_start_abs_time(original);
	BA_ASSERT_REAL_LESS_THAN_OR_EQUAL(universe_start_abs_time,
					  start_abs_time);

	// make sure original doesn't end after universe
	end_abs_time = ba_time_periods_last_end_abs_time(original);
	BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(universe_end_abs_time,
					     end_abs_time);
    }

    complement = ba_time_periods_new();

    current_abs_time = universe_start_abs_time;
    for(elem = original->periods; elem; elem = g_slist_next(elem)) {
	time_period = (ba_time_period_t *)elem->data;
	BA_ASSERT(time_period);
	start_abs_time = ba_time_period_start_abs_time_get(time_period);
	end_abs_time = ba_time_period_end_abs_time_get(time_period);

	// we don't use BA_ASSERT_REAL_LESS_THAN_OR_EQUAL() because we
	// strictly want to look for current_abs_time less than
	// start_abs time.  if we used the former, we'd create lots of
	// 0-length periods.
	if(current_abs_time < start_abs_time) {
	    append_start_end(complement, current_abs_time, start_abs_time);
	}
	current_abs_time = end_abs_time;
    }
    if(current_abs_time < universe_end_abs_time) {
	append_start_end(complement, current_abs_time, universe_end_abs_time);
    }

    if(simplify) {
	simplify_periods(complement); // can't hurt
    }
    return complement;
}

extern void
ba_time_periods_complement(ba_time_periods_t *a_time_periods,
			   ba_time_period_t *b_time_period,
			   ba_time_periods_t *result_time_periods)
{
    BA_ASSERT(a_time_periods && b_time_period && result_time_periods);
    ba_time_periods_t *tmp_time_periods;
    
    tmp_time_periods = complement_new_opt_simplify(a_time_periods,
						   b_time_period, BA_TRUE);
    
    swap_periods(tmp_time_periods, result_time_periods);
    ba_time_periods_delete(tmp_time_periods);
}

// append ////////////////////////////////////////////////////////////

static void
append_period_opt_copy_opt_simplify(ba_time_periods_t *time_periods,
				    ba_time_period_t *time_period,
				    ba_bool copy,
				    ba_bool simplify)
{
    BA_ASSERT(time_periods && time_period);
    GSList *elem;
    ba_time_period_t *insert_time_period;

    if(ba_time_periods_is_empty(time_periods) == BA_FALSE) {
	ba_time_period_t *last_time_period;
	ba_real start_abs_time, last_end_abs_time;
	
	elem = g_slist_last(time_periods->periods);
	BA_ASSERT(elem);
	last_time_period = (ba_time_period_t *)elem->data;
	last_end_abs_time = ba_time_period_end_abs_time_get(last_time_period);
	start_abs_time = ba_time_period_start_abs_time_get(time_period);
	BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(start_abs_time,
					     last_end_abs_time);
    }

    if(copy) {
	insert_time_period = ba_time_period_copy(time_period);
    } else {
	insert_time_period = time_period;
    }
    time_periods->periods = g_slist_append(time_periods->periods,
					   insert_time_period);

    if(simplify) {
	simplify_periods(time_periods);
    }
}

static void
append_start_end(ba_time_periods_t *time_periods,
		 ba_real start_abs_time,
		 ba_real end_abs_time)
{
    BA_ASSERT(time_periods);
    ba_time_period_t *time_period;

    time_period = ba_time_period_new(start_abs_time, end_abs_time);
    append_period_opt_copy_opt_simplify(time_periods, time_period, BA_FALSE,
					BA_FALSE);
}

extern void
ba_time_periods_append_period(ba_time_periods_t *time_periods,
			      ba_time_period_t *time_period)
{
    BA_ASSERT(time_periods && time_period);
    append_period_opt_copy_opt_simplify(time_periods, time_period, BA_TRUE,
					BA_TRUE);
}

extern void
ba_time_periods_append_start_end(ba_time_periods_t *time_periods,
				 ba_real start_abs_time,
				 ba_real end_abs_time)
{
    BA_ASSERT(time_periods);
    ba_time_period_t *time_period;

    time_period = ba_time_period_new(start_abs_time, end_abs_time);
    append_period_opt_copy_opt_simplify(time_periods, time_period, BA_FALSE,
					BA_TRUE);
}

// class stuff ///////////////////////////////////////////////////////

// be wary if i add stuff to ba_periods_t
extern ba_time_periods_t *
ba_time_periods_copy(ba_time_periods_t *time_periods_src)
{
    BA_ASSERT(time_periods_src);
    ba_time_periods_t *time_periods_dst;
    GSList *elem;
    ba_time_period_t *time_period_src;

    time_periods_dst = ba_time_periods_new();

    // could have used g_slist_foreach...
    for(elem = time_periods_src->periods; elem; elem = g_slist_next(elem)) {
	time_period_src = (ba_time_period_t *)elem->data;

	append_period_opt_copy_opt_simplify(time_periods_dst, time_period_src,
					    BA_TRUE, BA_FALSE);
    }

    return time_periods_dst;
}

extern ba_time_periods_t *
ba_time_periods_new(void)
{
    ba_time_periods_t *time_periods;

    time_periods = ba_mem_malloc(sizeof(ba_time_periods_t));

    time_periods->periods = 0;

    return time_periods;
}

extern void
ba_time_periods_delete(ba_time_periods_t *time_periods)
{
    BA_ASSERT(time_periods);

    g_slist_foreach(time_periods->periods, ba_time_period_delete_from_list, 0);
    g_slist_free(time_periods->periods);

    ba_mem_free(time_periods);
}

extern void
ba_time_periods_delete_from_list(void *time_periods_in,
				 void *unused)
{
    BA_ASSERT(time_periods_in);
    ba_time_periods_t *time_periods;
    
    time_periods = (ba_time_periods_t *)time_periods_in;
    
    ba_time_periods_delete(time_periods);
}

//////////////////////////////////////////////////////////////////////
