#ifndef __BA_TIME_PERIODS_H__
#define __BA_TIME_PERIODS_H__

#include <ba_config.h>

typedef struct ba_time_periods_s ba_time_periods_t;

extern void ba_time_periods_nullify(ba_time_periods_t *time_periods);
extern ba_bool ba_time_periods_is_equal(ba_time_periods_t *a_time_periods,
					ba_time_periods_t *b_time_periods);
extern ba_bool ba_time_periods_is_equal_period(ba_time_periods_t
					       *a_time_periods,
					       ba_time_period_t
					       *b_time_period);
extern ba_bool ba_time_periods_is_equal_start_end(ba_time_periods_t
						  *a_time_periods,
						  ba_real start_abs_time,
						  ba_real end_abs_time);
extern ba_bool ba_time_periods_is_empty(ba_time_periods_t *time_periods);
extern ba_real ba_time_periods_total_delta_time(ba_time_periods_t
						*time_periods);
extern void ba_time_periods_print(ba_time_periods_t *time_periods);
extern ba_real ba_time_periods_first_start_abs_time(ba_time_periods_t
						    *time_periods);
extern ba_real ba_time_periods_last_end_abs_time(ba_time_periods_t
						 *time_periods);

extern void ba_time_periods_union(ba_time_periods_t *a_time_periods,
				  ba_time_periods_t *b_time_periods,
				  ba_time_periods_t *result_time_periods);
extern void ba_time_periods_union_period(ba_time_periods_t *a_time_periods,
					 ba_time_period_t *b_time_period,
					 ba_time_periods_t
					 *result_time_periods);
extern void ba_time_periods_union_start_end(ba_time_periods_t *a_time_periods,
					    ba_real start_abs_time,
					    ba_real end_abs_time,
					    ba_time_periods_t
					    *result_time_periods);

extern void ba_time_periods_intersection(ba_time_periods_t *a_time_periods,
					 ba_time_periods_t *b_time_periods,
					 ba_time_periods_t
					 *result_time_periods);
extern void ba_time_periods_intersection_period(ba_time_periods_t
						*a_time_periods,
						ba_time_period_t
						*b_time_period,
						ba_time_periods_t
						*result_time_periods);
extern void ba_time_periods_intersection_start_end(ba_time_periods_t
						   *a_time_periods,
						   ba_real start_abs_time,
						   ba_real end_abs_time,
						   ba_time_periods_t
						   *result_time_periods);

extern void ba_time_periods_difference(ba_time_periods_t *a_time_periods,
				       ba_time_periods_t *b_time_periods,
				       ba_time_periods_t *result_time_periods);
extern void ba_time_periods_difference_period(ba_time_periods_t
					      *a_time_periods,
					      ba_time_period_t
					      *b_time_period,
					      ba_time_periods_t
					      *result_time_periods);
extern void ba_time_periods_difference_period_r(ba_time_period_t
						*a_time_period,
						ba_time_periods_t
						*b_time_periods,
						ba_time_periods_t
						*result_time_periods);
extern void ba_time_periods_difference_start_end(ba_time_periods_t
						 *a_time_periods,
						 ba_real start_abs_time,
						 ba_real end_abs_time,
						 ba_time_periods_t
						 *result_time_periods);
extern void ba_time_periods_difference_start_end_r(ba_real start_abs_time,
						   ba_real end_abs_time,
						   ba_time_periods_t
						   *b_time_periods,
						   ba_time_periods_t
						   *result_time_periods);

extern void ba_time_periods_complement(ba_time_periods_t *a_time_periods,
				       ba_time_period_t *b_time_period,
				       ba_time_periods_t *result_time_periods);

extern void ba_time_periods_append_period(ba_time_periods_t *time_periods,
					  ba_time_period_t *time_period);
extern void ba_time_periods_append_start_end(ba_time_periods_t *time_periods,
					     ba_real start_abs_time,
					     ba_real end_abs_time);

extern ba_time_periods_t *ba_time_periods_copy(ba_time_periods_t
					       *time_periods_src);
extern ba_time_periods_t *ba_time_periods_new(void);
extern void ba_time_periods_delete(ba_time_periods_t *time_periods);
extern void ba_time_periods_delete_from_list(void *time_periods_in,
					     void *unused);

#endif /* !__BA_TIME_PERIODS_H__ */
